# -*- coding: utf-8 -*-
"""
Created on Thu Mar 10 21:28:51 2016

@author: Tobias Jachowski
"""
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.widgets import Cursor
from scipy.optimize import fsolve

from pyoti.modification.modification import Modification, GraphicalMod
from pyoti import traces as tc


class ITouchdown(GraphicalMod):
    """
    Subclass of Touchdown that provides graphical interfaces to adjust the fit
    parameters.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Define attributes needed for the graphical fitting of the touchdown
        # a cursor on the axes to show upper fitting border and left/right
        # (middle) border
        self._cursor = None
        self._lines = {}
        self._ax = None
        self.supertitle = None

    def _figure(self):
        """
        Initialize and show an interactive plot for fitting the touchdown.
        Adjust the fit parameters interactively by displaying the fitted
        touchdown. The plot of the fitting is stored in self.figure.
        """
        # create new figure and axes for fitting the touchdown
        figure, ax = plt.subplots(1, 1, sharex=True, sharey=True)
        self._ax = ax

        # create lines to plot the data for the fit and for displaying fitting
        # result
        self._lines['orig_data'] = ax.plot([0], [0], tc.color('psdZ'))[0]
        self._lines['binned_data'] = ax.plot([0], [0], 'b.')[0]

        # cursor to display the "boundaries" for fitdata
        self._cursor = Cursor(ax, useblit=True, color='red', lw=1)

        # initialize all lines for the plot that represent the fit paramters
        # extrapolation of fit left
        self._lines['ext_left'] = ax.plot([0, 0], [0, 0], 'c', lw=1)[0]
        # extrapolation of fit right
        self._lines['ext_right'] = ax.plot([0, 0], [0, 0], 'c', lw=1)[0]
        # fit left
        self._lines['fit_left'] = ax.plot([0, 0], [0, 0], 'r', lw=3)[0]
        # fit right
        self._lines['fit_right'] = ax.plot([0, 0], [0, 0], 'g', lw=3)[0]
        # touchdown position
        self._lines['touchdown'] = ax.axvline(x=0.0, color='y', lw=2)
        # touchdown of bead, where positionZ equal to 0.0
        self._lines['touchdown_calib'] = ax.axvline(x=0.0, color='k', lw=2,
                                                    dashes=(10, 10))

        ax.set_xlabel(tc.label('positionZ'))
        ax.set_ylabel(tc.label('psdZ'))
        ax.ticklabel_format(useOffset=False)
        ax.grid(True)

        self.supertitle = figure.suptitle("Fit touchdown determination (move "
                                          "to upper and right/left limit, "
                                          "left/right click, repeat ...)")

        # method to update plot and perform the fit
        figure.canvas.mpl_connect('button_release_event',
                                  self._update_touchdown)

        return figure

    def _update_fig(self, data_changed=True, **kwargs):
        """
        Updates the interactive fitting plot: Set values of fit parameters,
        touchdown position and fitted lines
        """
        if data_changed:
            # data, [:,0] is positionZ, [:,1] is psdZ
            data = self.modification._get_data_based(
                traces=['positionZ', 'psdZ'], decimate=True, copy=False)
            self.modification.validate_fit_params(data=data)
            bin_means = self.modification.bin_means
            minx = data[:, 0].min() - 0.05e-6  # m
            maxx = data[:, 0].max() + 0.05e-6  # m
            miny = data[:, 1].min() - 0.05
            maxy = data[:, 1].max() + 0.05
            self._ax.set_xlim([minx * 1e6, maxx * 1e6])  # µm
            self._ax.set_ylim([miny, maxy])

            self._lines['orig_data'].set_data(data[:, 0] * 1e6, data[:, 1])
            self._lines['binned_data'].set_data(bin_means[:, 0] * 1e6,
                                                bin_means[:, 1])

        # get left and right datapoint means according to fit borders for
        # fitting
        bin_means = self.modification.bin_means
        left_means = self.modification.left_means
        right_means = self.modification.right_means

        self._lines['ext_left'].set_data(bin_means[:, 0] * 1e6,
                                         self.get_pv_left(bin_means[:, 0]))
        self._lines['ext_right'].set_data(bin_means[:, 0] * 1e6,
                                          self.get_pv_right(bin_means[:, 0]))
        self._lines['fit_left'].set_data(left_means[:, 0] * 1e6,
                                         self.get_pv_left(left_means[:, 0]))
        self._lines['fit_right'].set_data(right_means[:, 0] * 1e6,
                                          self.get_pv_right(right_means[:, 0]))
        self._lines['touchdown'].set_xdata([self.modification.touchdown * 1e6,
                                            self.modification.touchdown * 1e6])
        self._lines['touchdown_calib'].set_xdata(
            [self.modification._calibration.touchdown * 1e6,
             self.modification._calibration.touchdown * 1e6])

    def _pre_close_fig(self):
        # Store touchdown fit plot for documentation
        self.supertitle.set_text("Touchdown determined by polynomial fits")
        self._cursor.visible = False
        # update visibility status of self._cursor before removing reference
        # to it
        self._figure_canvas_draw()

        # Release unused memory
        self._cursor = None
        self._ax = None
        self._lines.clear()
        self.supertitle = None

    def _update_touchdown(self, event):
        """
        Gets called by events of the canvas object of the figure
        """
        self._set_fit_params(event)
        touchdown = self.modification.fit_touchdown()
        self.modification._set_touchdown_leave_auto(touchdown)
        self.update_fig(data_changed=False)

    def _set_fit_params(self, event):
        """
        Set the parameters for the fitting routine.
        """
        if event.button == 1:  # left mouse button -> adjust left fitdata
            # left border for fitdata
            self.modification.left = self._cursor.linev.get_xdata()[0] * 1e-6  # m
            # max value (psdZ) for fitdata
            self.modification.left_upper = self._cursor.lineh.get_ydata()[0]
        else:  # right (or middle) mouse button -> adjust right fitdata
            # right border for fitdata
            self.modification.right = self._cursor.linev.get_xdata()[0] * 1e-6  # m
            # max value (psdZ) for fitdata
            self.modification.right_upper = self._cursor.lineh.get_ydata()[0]

    def get_pv_left(self, x):
        return np.polyval(self.modification.pf_left, x)

    def get_pv_right(self, x):
        return np.polyval(self.modification.pf_right, x)


class Touchdown(Modification):
    """
    Determine the approximate touchdown
    The displacement sensitivity and stiffness of a calibration are always
    relative to and given absolute at the touchdown of the bead on the surface
    The positionZ is touchdown corrected by this modification.
    """
    GRAPHICALMOD = ITouchdown

    def __init__(self, fit_touchdown=True, **kwargs):
        traces_apply = ['positionZ']
        # the touchdown position determined by calibration and fitting
        super().__init__(automatic_switch=True, datapoints=2500,
                         traces_apply=traces_apply, **kwargs)

        # register a widget, giving a key, a function called upon change
        self.add_iattribute('touchdown', description="Touchdown (m)",
                            value=0.0)

        # Initially set touchdown to touchdown of calibration, if touchdown
        # should not be automatically fitted. This will automatically switch of
        # the automatic fitting.
        if not fit_touchdown:
            self.touchdown = self._calibration.touchdown

        # register a button to set the touchdown from calibration.touchdown
        self.add_iattribute('set_touchdown_calib',
                            description="Set touchdown from calibration",
                            callback_functions=[
                                self._set_touchdown_from_calibration,
                                self.evaluate])

        # Define attributes needed for the fitting of the touchdown
        self.left_upper = None  # max value (psdZ) for fitdata
        self.right_upper = None  # max value (psdZ) for fitdata
        self.left = None  # right/left border for fitdata
        self.right = None  # right/left border for fitdata

    def _recalculate(self):
        # calculate data for fitting
        self.validate_fit_params()

        # fit the touchdown
        touchdown = self.fit_touchdown()
        self._set_touchdown_leave_auto(touchdown)

    def _modify(self, data, samples, data_traces, data_index, mod_index):
        data[:, data_index] -= self.touchdown
        return data

    def validate_fit_params(self, data=None):

        # self.init_fit_parameters()
        if data is None:
            data = self._get_data_based(traces=['positionZ', 'psdZ'],
                                        decimate=True, copy=False)

        positionzmin = data[:, 0].min()
        positionzmax = data[:, 0].max()
        middle = positionzmin + (positionzmax - positionzmin) / 2
        upper = data[:, 1].max()

        # First initialisation of fit boundaries
        if self.left is None or self.right_upper is None:
            self.left_upper = upper   # max value (psdZ) for fitdata
            self.left = middle        # right/left border for fitdata
        if self.right is None or self.right_upper is None:
            self.right = middle       # right/left border for fitdata
            self.right_upper = upper  # max value (psdZ) for fitdata

        bin_means, bin_width = self.calculate_bin_means(data=data)
        self.bin_means = bin_means

        if not np.any(self.left_means_idx):
            # first, change only the upper boundary and leave the left one
            # usually solves problems, when psdZ was changed by an offset
            self.left_upper = upper
            if not np.any(self.left_means_idx):
                # still no means, change left boundary, too
                self.left = middle
        if not np.any(self.right_means_idx):
            # first, change only the upper boundary and leave the right one
            # usually solves problems, when psdZ was changed by an offset
            self.right_upper = upper
            if not np.any(self.right_means_idx):
                # still no means, change right boundary, too
                self.right = middle

    def fit_touchdown(self):
        """
        Performs the fit.
        In essence, it fits 2 polynomials of 2nd degree to the left (bead
        on surface) and to the right (free bead) portion of the psdZ data
        The intercept of these two polynomials determine the touchdown of the
        bead with the surface.
        """
        if self.left_means.shape[0] > 0 and self.right_means.shape[0] > 0:
            def f_left(x):
                return np.polyval(self.pf_left, x)

            def f_right(x):
                return np.polyval(self.pf_right, x)

            def findIntersection(fun0, fun1, x0):
                # one could also solve formula for the intersection
                # analytically, but fsolve is more flexible
                return fsolve(lambda x: fun0(x) - fun1(x), x0)
            middle = self.left + (self.right - self.left)/2
            touchdown = findIntersection(f_left, f_right, middle)[0]

            return touchdown

    def _set_touchdown_leave_auto(self, touchdown):
        self.iattributes.set_value('touchdown', touchdown,
                                   leave_automatic=True)

    def _set_touchdown_from_calibration(self):
        self.touchdown = self._calibration.touchdown

    @property
    def touchdown(self):
        return self.iattributes.touchdown

    @touchdown.setter
    def touchdown(self, touchdown):
        self.iattributes.touchdown = touchdown

    @property
    def _calibration(self):
        # calibration, which is used to:
        # - set touchdown from calibration
        # - visualize calibration.touchdown in GTouchdown
        return self.view_apply.calibration

    @property
    def pf_left(self):
        # get left datapoint means according to fit borders for fitting
        left_means = self.left_means
        # polynomial 2nd order order for negative positionZ
        return np.polyfit(left_means[:, 0], left_means[:, 1], 2)

    @property
    def pf_right(self):
        # get right datapoint means according to fit borders for fitting
        right_means = self.right_means
        # polynomial 2nd order for positive positionZ
        return np.polyfit(right_means[:, 0], right_means[:, 1], 2)

    @property
    def left_means(self):
        return self.bin_means[self.left_means_idx]

    @property
    def right_means(self):
        return self.bin_means[self.right_means_idx]

    @property
    def left_means_idx(self):
        x_check = self.bin_means[:, 0] < self.left
        y_check = self.bin_means[:, 1] < self.left_upper
        return np.logical_and(x_check, y_check)

    @property
    def right_means_idx(self):
        x_check = self.bin_means[:, 0] > self.right
        y_check = self.bin_means[:, 1] < self.right_upper
        return np.logical_and(x_check, y_check)

    @property
    def bin_means(self):
        if not hasattr(self, '_v_bin_means'):
            return None
        else:
            return self._v_bin_means

    @bin_means.setter
    def bin_means(self, bin_means):
        self._v_bin_means = bin_means



# -*- coding: utf-8 -*-
# """
# - Author: steve simmert
# - E-mail: steve.simmert@uni-tuebingen.de
# - Copyright: 2015
# """
__author__ = "Steve Simmert"
__copyright__ = "Copyright 2016, The PyOTIC Project"
__credits__ = []
__license__ = "Apache-2.0"
__maintainer__ = "Steve Simmert"
__email__ = "steve.simmert@uni-tuebingen.de"
__status__ = "stable"

import sys
import os

directory = os.path.dirname(globals()['__file__'])

try:
    with open(os.path.join(directory, 'VERSION.txt')) as f:
        __version__ = f.read().strip()
except:
    with open(os.path.join(directory, '..', 'VERSION.txt')) as f:
        __version__ = f.read().strip()


def version():
    """
    Returns the version.
    """
    return __version__


from IPython import get_ipython
import ipykernel

#Load matplotlib and set backend:
if 'IPython' in sys.modules:
    # We are in an ipython/juypter environment
    ip = get_ipython()
    if 'ZMQInteractiveShell' in repr(ip):
        # We are in a notebook or jupyter lab,
        # try to use ipympl backend (module://ipympl.backend_nbagg)
        try:
            import matplotlib
            matplotlib.use('module://ipympl.backend_nbagg')
        except ImportError:
            pass

    # Set format for inline plots
    from IPython.display import set_matplotlib_formats
    # %config InlineBackend.figure_formats = ['png']
    set_matplotlib_formats('png', 'svg', 'pdf', 'jpeg', quality=90)


import matplotlib.pyplot as plt

# load pint and create unit registry
from pint import UnitRegistry
ureg = UnitRegistry()

from . import focal_shift

from .height_calibration import gen_height_fit_pars
from .height_calibration import fit_rel_drag
from .height_calibration import fit_height_data
from .height_calibration import HeightCalibration
from .height_calibration import HeightFitResult
from .height_calibration import HeightCalibTime

from . import name_constants as co

from .physics import drag
from .physics import faxen_factor

from . import plotting
from .plotting import add_plot_to_figure

from . import psd
from .psd import calculate_psd
from .psd import ExpSetting
from .psd import gen_PSD_from_time_series
from .psd import PSD
from .psd import PSDMeasurement
from .psd import gen_psdm_from_region

from . import psd_fitting
from .psd_fitting import gen_psd_fit_pars
from .psd_fitting import PSDFit

from . import utilities

from scipy.constants import Boltzmann as k_B



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 11 15:32:58 2016

@author: Tobias Jachowski
"""
import numpy as np
import pint

from pyoti import config as cfg
from pyoti.calibration.calibsource import CalibrationSource
from pyoti import helpers as hp


class PyOTICfSource(CalibrationSource):
    def __init__(self, filename=None, focalshift=None, name=None, **kwargs):
        if filename is None:
            raise TypeError("PyOTICfSource missing the required positional "
                            "argument 'filename'.")
        self.filename = filename

        value, factor = read_pyotic_hc_result(filename)

        self.corrfactor = value['correction_factor']
        self.dsurf = value['apparent_surface_distance'] \
            * factor['apparent_distance']
        self.beta = np.array([value['beta_x'],
                              value['beta_y'],
                              value['beta_z']]) * factor['beta']
        self.kappa = np.array([value['kappa_x'],
                               value['kappa_y'],
                               value['kappa_z']]) * factor['kappa']
        self.mbeta = np.array([value['mbeta_x'],
                               value['mbeta_y'],
                               value['mbeta_z']]) * factor['mbeta']
        self.mkappa = np.array([value['mkappa_x'],
                                value['mkappa_y'],
                                value['mkappa_z']]) * factor['mkappa']
        self.radiusspec = value['radius'] * factor['radius']
        self.focalshift = focalshift or value['focal_shift']
        self.name = name or 'OT Investigator and Calibrator Height ' \
                            'Calibration file originally loaded from \n    ' \
                            '%s' % (self.filename)


def read_pyotic_hc_result(filename):
    pars = cfg.read_cfg_file(filename)

    ureg = pint.UnitRegistry()

    results = pars['RESULTS']
    units = pars['UNITS']

    values = {}
    for key, value in results.items():
        try:
            values[key] = float(value)
        except:
            values[key] = value

    factors = {}
    factors['radius'] = ureg(units['radius']).to('m').magnitude
    factors['apparent_distance'] \
        = ureg(units['apparent_distance']).to('m').magnitude
    factors['beta'] = ureg(units['beta']).to('m/V').magnitude
    factors['mbeta'] = ureg(units['mbeta']).to('m/(V*m)').magnitude
    factors['kappa'] = ureg(units['kappa']).to('N/m').magnitude
    factors['mkappa'] = ureg(units['mkappa']).to('N/(m*m)').magnitude

    return values, factors


class PyOTICSource(CalibrationSource):
    def __init__(self, heightcalibration, focalshift=0.785, name=None,
                 **kwargs):
        self.hc = heightcalibration

        self.name = name or 'OT Investigator and Calibrator Height Calibration'

    @property
    def corrfactor(self):
        return self.hc.height_fit_results.corr
        # value['correction_factor']

    @property
    def dsurf(self):
        return self.hc.height_fit_results.h0
        # value['apparent_surface_distance']
        # * ureg(units['apparent_distance']).to('m').magnitude

    @property
    def beta(self):
        hfr = self.hc.height_fit_results
        return np.array([hfr.beta_x, hfr.beta_y, hfr.beta_z])
        # * ureg(units['beta']).to('m/V').magnitude

    @property
    def kappa(self):
        hfr = self.hc.height_fit_results
        return np.array([hfr.kappa_x, hfr.kappa_y, hfr.kappa_z])
        # * ureg(units['kappa']).to('N/m').magnitude

    @property
    def mbeta(self):
        hfr = self.hc.height_fit_results
        return np.array([hfr.mbeta_x, hfr.mbeta_y, hfr.mbeta_z])
        # * ureg(units['mbeta']).to('m/(V*m)').magnitude

    @property
    def mkappa(self):
        hfr = self.hc.height_fit_results
        return np.array([hfr.mkappa_x, hfr.mkappa_y, hfr.mkappa_z])
        # * ureg(units['mkappa']).to('N/(m*m)').magnitude

    @property
    def radiusspec(self):
        return self.hc.height_fit_results.radius
        # value['radius']
        # * ureg(units['radius']).to('m').magnitude

    @property
    def focalshift(self):
        return self.hc.focal_shift


class PyOTICSingleSource(CalibrationSource):
    def __init__(self, filename=None, number=1, active=True, radiusspec=None,
                 corrfactor=1.0, focalshift=None, mbeta=None, mkappa=None,
                 dsurf=0.0, name=None, **kwargs):
        if filename is None:
            raise TypeError("PyOTICSingle missing the required positional "
                            "argument 'filename'.")
        self.filename = filename
        self.number = 1
        self.active = active

        # read in the calibration data and parameters
        data, pars = read_pyotic_single_calibration(filename, number=number)

        # calculate conversion factors
        ureg = pint.UnitRegistry()
        factor = {}
        factor['radius'] = ureg(pars['radius_unit']).to('m').magnitude
        factor['beta'] \
            = ureg(pars['displacement_sensitivity_unit']).to('m/V').magnitude
        factor['kappa'] \
            = ureg(pars['trapstiffness_unit']).to('N/m').magnitude

        # extract the calibration data
        self.excited_axis = int(data['ex_axis'])
        beta_x_pc = data['beta_x_pc'] * factor['beta']
        beta_y_pc = data['beta_y_pc'] * factor['beta']
        beta_z_pc = data['beta_z_pc'] * factor['beta']
        self.beta_pc = np.array([beta_x_pc, beta_y_pc, beta_z_pc])
        beta_x_ac = data['beta_x_ac'] * factor['beta']
        beta_y_ac = data['beta_y_ac'] * factor['beta']
        beta_z_ac = data['beta_z_ac'] * factor['beta']
        self.beta_ac = np.array([beta_x_ac, beta_y_ac, beta_z_ac])
        kappa_x_pc = data['kappa_x_pc'] * factor['kappa']
        kappa_y_pc = data['kappa_y_pc'] * factor['kappa']
        kappa_z_pc = data['kappa_z_pc'] * factor['kappa']
        self.kappa_pc = np.array([kappa_x_pc, kappa_y_pc, kappa_z_pc])
        kappa_x_ac = data['kappa_x_ac'] * factor['kappa']
        kappa_y_ac = data['kappa_y_ac'] * factor['kappa']
        kappa_z_ac = data['kappa_z_ac'] * factor['kappa']
        self.kappa_ac = np.array([kappa_x_ac, kappa_y_ac, kappa_z_ac])
        if mbeta is None:
            mbeta = [0.0 for i in range(self.beta.size)]
        self.mbeta = np.array(hp.listify(mbeta))
        if mkappa is None:
            mkappa = [0.0 for i in range(self.kappa.size)]
        self.mkappa = np.array(hp.listify(mkappa))
        self.dsurf = dsurf
        self.radiusspec = radiusspec \
            or pars['specified_radius'] * factor['radius']
        self.corrfactor = corrfactor
        self.focalshift = focalshift or pars['focal_shift']
        self.name = name or 'OT Investigator and Calibrator single ' \
                            'calibration originally loaded from \n    %s' \
                            % (self.filename)

    @property
    def beta(self):
        if self.active:
            return self.beta_ac
        else:
            return self.beta_pc

    @beta.setter
    def beta(self, beta):
        print('Setting beta not supported. Instead, set the property '
              '`beta_pc` or `beta_ac`.')

    @property
    def kappa(self):
        if self.active:
            return self.kappa_ac
        else:
            return self.beta_pc

    @kappa.setter
    def kappa(self, kappa):
        print('Setting kappa not supported. Instead, set the property '
              '`kappa_pc` or `kappa_ac`.')


def read_pyotic_single_calibration(hc_datafile, hc_parafile=None, number=1):
    # Autodetermine the name of the parameterfile
    hc_parafile = hc_parafile or hc_datafile.replace('_data.dat',
                                                     '_parameters.txt')

    # read in the parameters of the calibration
    pars = cfg.read_cfg_file(hc_parafile)
    # convert strings into floats
    parameters = {}
    for key, value in pars['DEFAULT'].items():
        try:
            parameters[key] = float(value)
        except:
            parameters[key] = value

    # read in the data of the first calibration
    with open(hc_datafile) as f:
        lines = f.readlines()
        header = lines[0].split()
        calibration = lines[number].split()
    # create a dict with all the data of the calibration
    data = dict(zip(header, [float(i) for i in calibration]))

    return data, parameters



# -*- coding: utf-8 -*-
"""
Created on Sun Mar  5 15:23:27 2017

@author: Tobias Jachowski
"""
import collections
import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import UnivariateSpline

from pyoti.modification.modification import Modification, GraphicalMod
from pyoti import helpers as hp
from pyoti import traces as tc
from pyoti.evaluate import signal as sn
from pyoti.evaluate.tether import Tether


class IBaseline(GraphicalMod):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _figure(self):
        # Create figure and axes to plot
        plot_params = {'wspace': 0.09, 'hspace': 0.1}
        self._set_plot_params(plot_params)
        fig, ax = plt.subplots(2, 2)
        ax = ax.flatten()

        # create lines to plot the data
        for i, trace in enumerate(self.modification.traces_apply):
            # original data
            ax[i].plot([0], [0], 'c')
            # corrected data
            ax[i].plot([0], [0], 'm')
            # plot the model
            ax[i].plot([0], [0], 'r')
            # datapoints, which were used to fit the model
            ax[i].plot([0], [0], 'r.', ms=3.5)

            ax[i].legend(('original', 'corrected', 'fit', 'fit points'),
                         loc='best')
            ax[i].set_xlabel("Time in (s)")
            ax[i].set_ylabel(tc.label(trace))
            ax[i].grid(True)

        fig.suptitle("Original and corrected data and datapoints used.")

        return fig

    def _update_fig(self, **kwargs):
        # get the axes to plot
        ax = self.figure.axes

        # plot with 200 Hz
        decimate = int(np.round(self.modification.view_based.samplingrate
                                / 200))
        # average with 1000 Hz
        window = int(np.round(self.modification.view_based.samplingrate
                              / 1000))

        # Get the data to be plotted
        t = self.modification.view_based.timevector[::decimate]
        data = self.modification._get_data_based(
            traces=self.modification.traces_apply, copy=False,
            decimate=decimate, window=window)
        t_f, data_f = self.modification._calculate_fit_data()

        # Plot the data
        for i, trace in enumerate(self.modification.traces_apply):
            # original data
            ax[i].lines[0].set_data(t, data[:, i])
            # corrected data
            ax[i].lines[1].set_data(t, data[:, i]
                                    - self.modification._model[trace](t))
            # plot the model
            ax[i].lines[2].set_data(t, self.modification._model[trace](t))
            # datapoints, which were used to fit the model
            ax[i].lines[3].set_data(t_f, data_f[:, i])

            # recompute ax.dataLim
            ax[i].relim()
            # update ax.viewLim using new dataLim
            ax[i].autoscale_view()

        # plot the results of the tether object
        ax[3].clear()
        tether = self.modification._create_tether()
        tether._rfigure(ax=ax[3])


class Baseline(Modification):
    """
    Corrects the baseline of psdX, psdY and, psdZ.

    This Modification's `view_based` and `view_apply` need to have the same
    parent Region!
    """
    GRAPHICALMOD = IBaseline

    def __init__(self, db_update=False, **kwargs):
        traces_apply = ['psdX', 'psdY', 'psdZ']
        super().__init__(traces_apply=traces_apply, automatic_switch=True,
                         **kwargs)

        description = "Automatically detect baseline index"
        self.add_iattribute('auto_detect_idx', description=description,
                            value=True, unset_automatic=False)

        # binning time to calculate the models in s
        description = "Time window (s) used to calculate the smoothing model"
        self.add_iattribute('bin_time', description=description,
                            value=0.01, unset_automatic=False)

        description = "Traces to be corrected"
        self.add_iattribute('modify_traces', description=description,
                            value=self.traces_apply, unset_automatic=False)

        # Initialize the baseline index to meaningfull values. The indices
        # mark the datpoints that are used to calculate the models.
        self._baseline_idx = None
        if not db_update:
            self._calculate_baseline_idx()

        # smoothing models for self.traces_apply (usually UnivariatSpline),
        # which are used to calculate "fitted" values for the baseline,
        # which are subtracted as correction from the traces.
        self._model = {}

    def _recalculate(self):
        # Update the baseline_idx which is used to calculate the means
        if self.iattributes.auto_detect_idx:
            self._calculate_baseline_idx()

        # Calculate the bin means for the data to be used
        t, means = self._calculate_fit_data()

        # Calculate the model for modifying the data
        self._calculate_model(t, means)

    def _calculate_baseline_idx(self):
        tether = self._create_tether()
        _idx = tether.baseline_idx(**self.baseline_kwargs)
        self._baseline_idx = tether.decimate_and_limit(_idx)

    def _create_tether(self):
        # Get tether arguments
        tether_kwargs = self.tether_kwargs
        return Tether(**tether_kwargs)

    def _calculate_fit_data(self):
        """
        Calculate the means of the baseline and return the times, where the
        means were calculated, and the means itself.
        """
        # Get the data to calculate the means from
        data = self._get_data_based(traces=self.traces_apply, copy=False)

        # Get the indices of where the bin means should be calculated from,
        # and take the datapoints around these indices with a length according
        # to self.bin_time
        resolution = self.view_based.samplingrate
        bin_time = self.iattributes.bin_time
        window = int(np.round(resolution * bin_time / 2))
        starts = self.baseline_idx - window
        stops = self.baseline_idx + window + 1
        segs = np.c_[starts, stops]
        max_stop = self.view_based.datapoints
        segs = sn.limit_segments(segs, max_stop=max_stop)
        slices = sn.idx_segments_to_slices(segs)

        # Get the indices of the time points which are used as x values for the
        # model fitting
        t = self.view_based.timevector[self.baseline_idx]

        # Finaly calculate the means
        means = calculate_means(data, slices)

        return t, means

    def _calculate_model(self, t, means):
        """
        Fit a model (UnivariatSpline) to the given data.

        t : 1D numpy.ndarray
        means : 2D numpy.ndarray
            The y data to calculate the model. The shape of `means` has to be
            (`t.size`, `len(self.traces_apply)`)
        """
        for i, trace in enumerate(self.traces_apply):
            y = means[:, i]
            us = UnivariateSpline(t, y, s=0, ext=1)
            self._model[trace] = us

        self.set_changed()

    def _modify(self, data, samples, data_traces, data_index, mod_index):
        # Get the shift of the time of the views the models were
        # calculated from (view_based) and the data is modified with
        # (view_apply). It is essential, that view_based and view_apply have
        # the same parent.
        shift = self.view_based.tmin - self.view_apply.tmin

        # Get the timevector for the requested samples and correct it with the
        # timeshift. It is used by the UnivariatSpline
        t = self.view_apply.timevector[samples]
        t -= shift

        # Get the indices of the data traces to be modified and the
        # names of the traces to be modified
        data_indices = hp.listify(data_index)
        mod_traces = np.array(self.traces_apply)[mod_index]

        # Modify the data with the UnivariatSpline
        for didx, mtrace in zip(data_indices, mod_traces):
            # Check if trace should be modified
            if mtrace in self.iattributes.modify_traces:
                data[:, didx] -= self._model[mtrace](t)

        return data

    @property
    def baseline_idx(self):
        return self._baseline_idx

    @baseline_idx.setter
    def baseline_idx(self, baseline_idx):
        self._baseline_idx = np.unique(baseline_idx)
        self.iattributes.auto_detect_idx = False
        self.set_changed()

    @property
    def tether_kwargs(self):
        if not hasattr(self, '_tether_kwargs') or self._tether_kwargs is None:
            tether_kwargs = {}
        else:
            tether_kwargs = self._tether_kwargs.copy()
        # Default argument `region` to view_based
        tether_kwargs['region'] = tether_kwargs.get('region', self.view_based)
        return tether_kwargs

    @tether_kwargs.setter
    def tether_kwargs(self, tether_kwargs):
        self._tether_kwargs = tether_kwargs
        self.set_changed()

    @property
    def baseline_kwargs(self):
        if not hasattr(self, '_baseline_kwargs') \
                or self._baseline_kwargs is None:
            return {}
        baseline_kwargs = self._baseline_kwargs.copy()
        return baseline_kwargs

    @baseline_kwargs.setter
    def baseline_kwargs(self, baseline_kwargs):
        self._baseline_kwargs = baseline_kwargs
        self.set_changed()


def calculate_means(data, samples=None, stds=False):
    """
    Calculate the means of `data`. If `samples_idx` is given, calculate the
    means of all samples contained in `samples_idx`.
    Additionally, the standard deviations can be returned, if stds is set to
    True.

    Parameters
    ----------
    data : numpy.ndarray
    samples : int, slice, or Iterable of int, slice, or index arrays
        Samples to calcualte the means from.
    stds : bool, optional
        Return also the standard deviations for the calculated means.

    Returns
    -------
    numpy.ndarray
        The calculated means
    (numpy.ndarray, numpy.ndarray)
        If `stds` is True, the calculated means and standard deviations.
    """
    if samples is None:
        samples = [slice(0, len(data))]
    if not isinstance(samples, collections.Iterable):
        samples = [samples]

    means = np.array([data[s].mean(axis=0) for s in samples], ndmin=2)

    if stds:
        stds = np.array([data[s].std(axis=0, ddof=1) for s in samples], ndmin=2)
        return means, stds

    return means


# The following is only to update to database version 0.8.0
class GBaseline(Baseline):
    pass



# -*- coding: utf-8 -*-
# """
# - Author: steve simmert
# - E-mail: steve.simmert@uni-tuebingen.de
# - Copyright: 2015
# """
"""
Manage power spectral densities measurements.
"""
import pdb

from . import ureg
from . import name_constants as co

from .data_parser import read_PSD_parameter_file
from .data_parser import read_std_data_file
from .data_parser import save_psd_data
from .data_parser import save_psd_params

from .physics import density_H2O
from .physics import drag
from .physics import faxen_factor
from .physics import MATERIALS
from .physics import Material
from .physics import oseen_factor
from .physics import viscosity_H2O
from .physics import dviscosity_H2O

from .plotting import add_plot_to_figure
from .plotting import col_dict

from .utilities import str2u

from collections import OrderedDict

import copy

from inspect import signature

import matplotlib.pyplot as plt

from os.path import join

from scipy import absolute
from scipy import array
from scipy import inf
from scipy import logical_or
from scipy import pi
from scipy import rand
from scipy import randn
from scipy import shape
from scipy import signal
from scipy import sqrt
from scipy import zeros

from scipy.fftpack import fft, ifft, fftfreq

import time

import warnings

#####--------------------------------------------------------------------------
#---- general  ---
#####--------------------------------------------------------------------------

# generate filtered random data
def gen_filtered_data(fun, fs, T_msr, *args, **kwargs):
    """
    Generate filtered random data.

    The function simulates data by calling a random number
    from a normal distribution, which simulates a trapped
    particle with pure white noise. The signal is Fourier-
    transformed and multiplied with the square-root of the given
    filter function. Thus, the filter function must represent the shape
    of a power spectral density. After that it's
    inverse-fourier-transformed.
    The number of data points is fs * T_msr, i.e. the sampling
    frequency times the total measurment time.

    To also account for the pyhsics, one can provide the following two
    keyword arguments:
    mean : float
        Set the mean of the output data.
    std : float
        Set the standard deviation of the output data.

    All other *args and **kwargs are passed to fun(*args, **kwargs).

    Arguments
    ---------
    fun : function
       Filter function, e.g. a low-pass filter.
    fs : float
        Sampling frequency
    T_msr : float
        Measurement time, the total time of the measurement.
    """
    mean = kwargs.pop('mean') if 'mean' in kwargs else 0.0
    std = kwargs.pop('std') if 'std' in kwargs else 1.0

    N = int(fs * T_msr)

    freq = fftfreq(N, 1/fs)

    xraw = randn(N)
    x = ifft(fft(xraw) * sqrt(fun(freq, *args, **kwargs))).real
    if std != 1.0:
        x_out = (x / x.std() * std)
    else:
        x_out = x

    x_out = x_out - x_out.mean() + mean

    return x_out


def calculate_psd(x, fs, N_win=1):
    """
    Calculate power spectral density of the signal x.

    N_win splits the signal into the corresponding number of parts.

    Arguments
    ---------
    x : array(float)
        Signal vector.
    fs : float
        Sampling frequency
    N_win : int
        Number of windows to devide the signal vector into. The power spectra
        are then averaged this many times.
    raise_exception : bool
        Raises an exception instead of a warning if N_win is no common divisor
        of the length of the signal vector.
    """
    N = len(x)
    rest = N % N_win
    if rest > 0:
        warnings.warn('N_win is no common divisor of N = len(x). '
                      '--> {0:1d} data points were omitted.'.format(rest))
        x = x[:-rest]  # throw away some data

    len_ = int(N / N_win)
    psds = []
    for idx in range(N_win):
        freq, psd = signal.welch(x[idx * len_: (idx + 1) * len_],
                                 fs=fs,
                                 window='boxcar',
                                 nperseg=len_,
                                 noverlap=0,
                                 nfft=None,
                                 detrend=None,
                                 return_onesided=True,
                                 scaling='density',
                                 axis=-1)
        psds.append(psd)

    return (freq, array(psds))


def gen_PSD_from_time_series(x, fs, N_win, calc_errors=False, **PSD_kwargs):
    """
    Generate a **psd** object from a time series data set.

    The function uses the Welch algorithm form the scipy.signal package.

    Arguments
    ---------
    x : array(float)
        Signal vector.
    fs : float
        Sampling frequency.
    N_win : int
        Number of windows to devide the signal vector into. The power spectra
        are then averaged this many times.
    calc_errors : bool
        If True, the errors of the psd values are determined from the single
        (windowed) psds. It is recommended to use the theoretical errors for
        PSD analysis (see Nørrelykke 2010), thus using the default: False.
    PSD_kwargs : keyword arguments
        passed over to the init call of **PSD**.

    Returns
    -------
    PSD
        Power spectral density as an object of the **PSD** class.
    """
    freq, psds = calculate_psd(x, fs, N_win=N_win)

    psd_avg = psds.mean(axis=0)
    if calc_errors:
        err = psds.std(axis=0) / sqrt(N_win)
    else:
        err = psd_avg / sqrt(N_win)

    p = PSD(freq,
            psd_avg,
            err=err,
            N_avg=N_win,
            f_sample=fs,
            **PSD_kwargs)

    return p


#####--------------------------------------------------------------------------
#---- objects ---
#####--------------------------------------------------------------------------


class PSD(object):
    def __init__(self,
                 freq,
                 psd,
                 err=None,
                 name='',
                 f_sample=None,
                 N_avg=1,
                 direction=None,
                 freq_unit='Hz',
                 psd_unit='V**2/Hz'
                 ):
        """
        Describes the one-sided power spectral density for one dimension.

        Arguments
        ---------
        freq : array(float)
            frequency vector of positive frequencies f in **Hertz**. This
            vector could still hold f=0. The psd-value psd(f=0) would then be
            taken as offset.
        psd : array(float)
            Vector with the corresponding power spectral densities. Normally,
            the qunatities are given in Volt²/Hz, since they are, e.g measured
            with a positional sensitive device. Thus the object assumes these
            units.
        err : array(float)
            Corresponding errors to the psd values with the same units.
        name : str
            Gives the PSD a name that is used to categorize it with other PSDs.
            This should idealy be one of 'x' 'y' or 'z' to ensure compatibility
            with HeightCalibration methods, but could be different for other
            purposes, e.g. collective_fit method of PSDFit object, where each
            psd must have a different name.
        f_sample : float
            Sampling frequency.
        N_avg : int
            Number of averages of the psd. In case no error vector is given
            this is  used to  calculate *err*.
        direction : str
            sould be either 'lateral' or 'axial' determining the direction
            in which the fluctuations are measured.
            'axial' means perpendicular to a sample surface.
        freq_unit : str
            Unit of the frequency values ('Hz').
        psd_unit : str
            Unit fo the psd values ('V**2/Hz').
        """
        self.N_avg = N_avg

        self._freq_unit = str2u(freq_unit)
        self._psd_unit = str2u(psd_unit)

        if 0 in freq:
            self.offset = psd[freq == 0][0]
        else:
            self.offset = 0.0

        self._freq = freq[freq > 0]
        self._psd = psd[freq > 0]

        if err is None:
            self._err = self._psd / sqrt(self.N_avg)
        else:
            self._err = err[freq > 0]

        self.reset_mask()

        self.name = name
        if direction is None:
            if 'z' in name:
                self.direction = 'axial'
            else:
                self.direction = 'lateral'
        else:
            self.direction = direction

        if f_sample is None:
            f_sample = 2 * max(freq)
        else:
            if 2 * max(freq) != f_sample:
                warnings.warn('Maximum frequency of the spectrum should, in '
                              'general, be half of the sampling frequency.\n'
                              '2 x f_max = {0:1.3f} != f_sample = {1:1.3f}.\n'
                              'You are able to continue, though.'
                              ''.format(2*max(freq), f_sample))
        self._f_sample = f_sample

    def _exclude_values(self, values, attr_name='freq'):
        """ Excludes values for attr == value """
        attr = getattr(self, attr_name)
        try:
            vals = iter(values)
        except:
            vals = [values]
        for val in vals:
            if val in attr:
                mask = attr == val
                self.mask = logical_or(self.mask, mask)
            else:
                warnings.warn('Value {} not available in attribute '
                              '{}.'.format(val, attr_name))

    def _exclude_values_outside(self, vmin, vmax, attr_name='freq'):
        """
        Exclude values outside the range vmin, vmax of attribute with name
        attr_name'.
        """
        attr = getattr(self, attr_name)
        mask = logical_or(attr < vmin, attr > vmax)
        self.mask = logical_or(self.mask, mask)

    @property
    def freq(self):
        return self._freq[~self.mask]

    def get_freq(self,
                 unit=None,
                 get_all=False,
                 get_masked=False,
                 offset=False):
        """
        Return the frequency vector.

        Arguments
        ---------
        unit : str
        get_all : bool
            If False uses the internal mask.
        get_masked : bool
            If True returns the masked instead of the unmasked elements.
        offset : bool
            Whether to include zero hertz to the freq-vector.
        """
        if unit is None or unit == self._freq_unit:
            conv = 1.0
        else:
            conv = ureg(self._freq_unit).to(unit).magnitude
        if get_masked:
            mask = self.mask
        else:
            mask = ~self.mask
        if get_all:
            f_out = self._freq * conv
        else:
            f_out = self._freq[mask] * conv
        if offset:
            a = list(f_out)
            a.insert(0, 0.0)
            f_out_off = array(a)
            return f_out_off
        else:
            return f_out

    @property
    def psd(self):
        return self._psd[~self.mask]

    def get_psd(self,
                unit=None,
                get_all=False,
                get_masked=False,
                offset=False):
        """
        Return the psd vector.

        Arguments
        ---------
        unit : str
        get_all : bool
            If False uses the internal mask.
        get_masked : bool
            If True returns the masked instead of the unmasked elements.
        offset : bool
            Whether to include the value at zero hertz.
        """
        if unit is None or unit == self._psd_unit:
            conv = 1.0
        else:
            conv = ureg(self._psd_unit).to(unit).magnitude
        if get_masked:
            mask = self.mask
        else:
            mask = ~self.mask
        if get_all:
            p_out = self._psd * conv
        else:
            p_out = self._psd[mask] * conv
        if offset:
            a = list(p_out)
            a.insert(0, self.offset * conv)
            p_out_off = array(a)
            return p_out_off
        else:
            return p_out

    @property
    def psd_err(self):
        return self._err[~self.mask]

    def get_err(self,
                unit=None,
                get_all=False,
                get_masked=False,
                offset=False):
        """
        Return the error vector.

        Arguments
        ---------
        unit : str
        get_all : bool
            If False uses the internal mask.
        get_masked : bool
            If True returns the masked instead of the unmasked elements.
        offset : bool
            Whether to include the error value at zero hertz.
        """
        if unit is None or unit == self._psd_unit:
            conv = 1.0
        else:
            conv = ureg(self._psd_unit).to(unit).magnitude
        if get_masked:
            mask = self.mask
        else:
            mask = ~self.mask
        if get_all:
            e_out = self._err * conv
        else:
            e_out = self._err[mask] * conv
        if offset:
            offset_err = self.offset * conv / sqrt(self.N_avg)
            a = list(e_out)
            a.insert(0, offset_err)
            e_out_off = array(a)
            return e_out_off
        else:
            return e_out

    @property
    def f_sample(self):
        return self._f_sample

    def get_f_sample(self, unit=None):
        if unit is None or unit == self._freq_unit:
            return self._f_sample
        else:
            conv = ureg(self._freq_unit).to(unit).magnitude
            return self._f_sample * conv

    def set_f_sample(self, f_sample, unit):
        if unit == self._f_sample:
            conv = 1.0
        else:
            conv = ureg(unit).to(self._freq_unit).magnitude
        self._f_sample = f_sample * conv

    @property
    def df(self):
        """ Frequency resolution of the power spectrum"""
        return self._freq[1] - self._freq[0]

    @property
    def T_msr(self):
        """ Measurement time in seconds """
        return (1 / min(self.get_freq(unit='Hz', get_all=True)))

    @property
    def N_samples(self):
        """ Number of samples """
        return (self.T_msr * self.f_sample)

    def is_lateral(self):
        """
        Whether the direction attribute is set to 'lateral' (True) or
        'axial' (False).
        """
        if self.direction == 'lateral':
            return True
        elif self.direction == 'axial':
            return False
        else:
            raise Exception('Unknown direction {}'.format(self.direction))

    def reset_mask(self):
        """ Rest the internal mask, so all values a taken into account. """
        self.mask = zeros(shape(self._freq)) > 0

    def add_mask(self, mask):
        """
        Logical OR with self.mask.
        """
        self.mask = logical_or(self.mask, mask)

    def exclude_freq(self, f_exclude):
        """
        Exclude the data at frequency or frequencies. f_exclude can be both a
        list of or a single float.
        """
        self._exclude_values(f_exclude, attr_name='_freq')

    def exclude_freq_outside(self, fmin, fmax):
        """
        Exclude the frequencies outside the interval [fmin, fmax].
        """
        self._exclude_values_outside(fmin, fmax, attr_name='_freq')

    def plot_psd(self,
                 axis=None,
                 plot_all=False,
                 plot_masked=False,
                 plot_errors=False,
                 **kwargs
                 ):
        """
        Plots the power spectral density.

        Arguments
        ---------
        axis : axis
            Axis to plot to.
        plot_all : bool
            If True, omits the internal mask.
        plot_masked : bool
            When True plots the masked values instead.
        plot_errors : bool
            Plot error bars as well.
        **kwargs
            key-word arguments handed over to plot()

        Returns
        -------
        figure
        """
        freq = self.get_freq(get_all=plot_all, get_masked=plot_masked)
        psd = self.get_psd(get_all=plot_all, get_masked=plot_masked)
        err = self.get_err(get_all=plot_all, get_masked=plot_masked)

        if plot_masked:
            if self.name in col_dict:
                col = col_dict['o' + self.name]
            else:
                col = 'gray'
            fmt = 'o'
            if 'markersize' not in kwargs:
                kwargs['markersize'] = 3
        else:
            if self.name in col_dict:
                col = col_dict[self.name]
            else:
                col = tuple(rand(3))
            fmt = '-'

        if not plot_errors:
            err = []

        if axis is None:
            fig = None
        else:
            fig = axis.figure

        if 'color' not in kwargs and self.name in col_dict.keys():
            kwargs['color'] = col
        if 'fmt' not in kwargs:
            kwargs['fmt'] = fmt
        if 'linewidth' not in kwargs:
            kwargs['linewidth'] = 1.0
        if 'alpha' not in kwargs:
            kwargs['alpha'] = 0.5
        if 'title' not in kwargs:
            kwargs['title'] = 'PSD {}'.format(self.name)
        if 'showLegend' not in kwargs:
            kwargs['showLegend'] = True
        if 'legend_kwargs' not in kwargs:
            lg_kws = {'loc': 3}
            kwargs['legend_kwargs'] = lg_kws
        if 'fontsize' not in kwargs:
            kwargs['fontsize'] = 16

        ax = add_plot_to_figure(fig,
                                freq,
                                psd,
                                yerr=err,
                                label=self.name,
                                axis=axis,
                                logplot=True,
                                **kwargs
                                )
        ax.grid(which='major')
        plt.setp(ax,
                 xlabel='Frequency (Hz)',
                 ylabel=r'PSD ($\mathsf{V^2/Hz}$)'
                 )

        return ax.figure


class ExpSetting(object):
    """
    Describes the experimental setting of a Measurment.
    """
    def __init__(self,
                 temp,
                 radius,
                 temp_err=0.0,
                 radius_err=0.0,
                 height=inf,
                 density_particle=None,
                 density_medium=None,
                 viscosity=None,
                 viscosity_err=0.0,
                 material='',
                 medium='water',
                 temp_unit='K',
                 radius_unit='m',
                 height_unit='m',
                 density_particle_unit='kg/m**3',
                 density_medium_unit='kg/m**3',
                 viscosity_unit='Pa*s',
                 warn=True):
        """
        Define the experimental setting.

        Arguments
        ---------
        temp : float
            Temperature at which the measurement was performed.
        radius : float
            Radius of the bead.
        temp_err : float
            Error of the Temperature value.
        radius_err : float
            Radius error.
        height : float
            Apperent distance between trapped particle and surface.
        density_particle : float
            The mass density of the bead. If this is provided, **material**
            will only name the material, but has no other effect.
        density_medium : float or callable.
            Density of the medium. If callable, then the function is called
            as density_medium(temp), with temp in Kelvin. Otherwise the float
            value is used.
        viscosity : float or callable
            Viscosity of the medium. If callable, then the function is called
            as viscosity(temp), with temp in Kelvin. Otherwise the float value
            is used.
        viscosity_err : float, callable or None
            If callable, the function is used as
            absolute(viscosity_err(temp) * temp_err), whit temp in Kelvin.
            Otherwise the float value is used as error in the viscosity
            estimate. If it's None, a known medium must be given.
        material : str
            Name of the particle material. E.g. 'polystyrene', 'titania',
            'silica'. If the name is none of these the particle density must be
            provided.
        temp_unit : str
            Unit of the Temperature ('K').
        radius_unit : str
            Unit of the radius values (e.g. 'um').
        height_unit : str
            Unit of the height.
        density_particle_unit : str
            The unit the density is provided in, e.g. 'kg/m**3'
        density_medium_unit : str
            The unit the density is provided in, e.g. 'kg/m**3'
        viscosity_unit : str
            Unit of the provided viscosity, e.g. 'Pa*s'.

        Notes
        -----
        material
            If material is a known material a set desnity_particle will not
            have an effect.
        copy
            Use copy() to make a duplicate of the setting.
        """
        # temperature
        self._qtemp = ureg.Measurement(temp, temp_err, temp_unit)

        # radius
        self._radius = radius
        self._radius_err = radius_err
        self._radius_unit = str2u(radius_unit)

        # height
        self._height = height
        self._height_unit = height_unit

        # material, density_material
        try:
            self.material = Material(material,
                                     density=density_particle,
                                     density_unit=density_particle_unit)
        except:
            if density_particle:
                if material == '':
                    material_name = 'Unknown material'
                else:
                    material_name = material
                self.material = Material(material_name,
                                         density=density_particle,
                                         density_unit=density_particle_unit)
            else:
                if warn:
                    warnings.warn('Unknown material or density not given. '
                                  'Using fallback density=1050 kg/m**3 '
                                  '(Polystyrene).')
                self.material = Material('PS')
        self.get_density_particle = self.material.get_density

        # medium
        if medium.lower() in ['water', 'h2o']:
            self._density_medium = density_H2O(self.get_temp(unit='K'))
            self._density_medium_unit = str2u('kg/m**3')
            self._viscosity = viscosity_H2O(self.get_temp(unit='K'))
            dv = absolute(dviscosity_H2O(self.get_temp(unit='K')) *
                          self.get_temp_err(unit='K'))
            self._viscosity_err = dv
            self._viscosity_unit = str2u('Pa*s')
            self.medium = 'water'
        else:
            if density_medium and viscosity:
                T = self.get_temp(unit='K')
                dT = self.get_temp_err(unit='K')
                try:
                    self._density_medium = density_medium(T)
                except:
                    self._density_medium = density_medium
                self._density_medium_unit = str2u(density_medium_unit)
                try:
                    self._viscosity = viscosity(T)
                except:
                    self._viscosity = viscosity
                try:
                    self._viscosity_err = absolute(viscosity_err(T) * dT)
                except:
                    self._viscosity_err = viscosity_err
                self._viscosity_unit = str2u(viscosity_unit)
                self.medium = medium
            else:
                raise Exception('You need to specify density and viscosity of '
                                'the medium "{}"'.format(medium))

    @property
    def _temp_unit(self):
        return str(self._qtemp.units)

    @property
    def temp(self):
        return self._qtemp.value.magnitude

    def get_temp(self, unit=None):
        if unit is None or unit == self._temp_unit:
            return self.temp
        else:
            return self._qtemp.to(unit).value.magnitude

    def set_temp(self, temp, unit):
        """ Set the Temperature of the measurement in specified units. """
        T_err = self._qtemp.error.magnitude
        if unit == self._temp_unit:
            self._qtemp = ureg.Measurement(temp, T_err, unit)
        else:
            T = ureg.Measurement(temp, T_err, unit)
            self._qtemp = T.to(self._temp_unit)

    @property
    def temp_err(self):
        return self._qtemp.error.magnitude

    def get_temp_err(self, unit=None):
        if unit is None or unit == self._temp_unit:
            return self.temp_err
        else:
            return self._qtemp.to(unit).error.magnitude

    def set_temp_err(self, T_err, unit):
        """ Set the Temperature of the measurement in specified units. """
        T = self.temp
        if unit == self._temp_unit:
            self._qtemp = ureg.Measurement(T, T_err, unit)
        else:
            T_ = ureg.Measurement(T, T_err, unit)
            self._qtemp = T_.to(self._temp_unit)

    @property
    def radius(self):
        return self._radius

    def get_radius(self, unit=None):
        if unit is None or unit == self._radius_unit:
            return self._radius
        else:
            conv = ureg(self._radius_unit).to(unit).magnitude
            return self._radius * conv

    def set_radius(self, radius, unit):
        """
        Set the radius of the used bead in the specified unit.

        Unit is then converted into the preset units.
        """
        if unit == self._radius_unit:
            conv = 1.0
        else:
            conv = ureg(unit).to(self._radius_unit).magnitude
        self._radius = radius * conv

    @property
    def radius_err(self):
        return self._radius_err

    def get_radius_err(self, unit=None):
        if unit is None or unit == self._radius_unit:
            return self._radius_err
        else:
            conv = ureg(self._radius_unit).to(unit).magnitude
            return self._radius_err * conv

    def set_radius_err(self, radius_err, unit):
        """
        Set the error of the radius of the used bead in the specified unit.

        Unit is then converted into the preset units.
        """
        if unit == self._radius_unit:
            conv = 1.0
        else:
            conv = ureg(unit).to(self._radius_unit).magnitude
        self._radius_err = radius_err * conv

    @property
    def height(self):
        """ Apperent distance between trapped particle and surface."""
        return self._height

    def get_height(self, unit=None):
        """
        Return the apperent distance between trapped particle and surface in
        the specified units.
        """
        if unit is None or unit == self._height_unit:
            return self._height
        else:
            conv = ureg(self._height_unit).to(unit).magnitude
            return self._height * conv

    def set_height(self, height, unit):
        """
        Set the apperent distance between trapped particle and surface in
        specified units.
        """
        conv = ureg(unit).to(self._height_unit).magnitude
        self._height = height * conv

    @property
    def density_particle(self):
        """Return the particle density in kg/m**3."""
        return self.material.get_density(unit='kg/m**3')

    #NOTE: get_density_particle is defined in __init__ and is hooked to
    #      material.get_density

    def set_material(self, name, density, density_unit):
        self.material = Material(name,
                                 density=density,
                                 density_unit=density_unit)

    @property
    def viscosity(self):
        return self._viscosity

    def get_viscosity(self, unit=None):
        if unit is None or unit == self._viscosity_unit:
            conv = 1.0
        else:
            conv = ureg(self._viscosity_unit).to(unit).magnitude
        return self._viscosity * conv

    def set_viscosity(self, viscosity, unit='Pa*s'):
        """
        Set the viscosity.

        Arguments
        ---------
        viscosity : float or callable
            Viscosity of the medium. If callable, then the function is called
            as viscosity(temp), with temp in Kelvin. Otherwise the float value
            is used.
        unit : str
            unit of the viscosity
        """
        T = self.get_temp(unit='K')
        try:
            self._viscosity = viscosity(T)
        except:
            self._viscosity = viscosity
        self._viscosity_unit = str2u(unit)

    def get_viscosity_err(self, unit=None):
        if unit is None or unit == self._viscosity_unit:
            conv = 1.0
        else:
            conv = ureg(self._viscosity_unit).to(unit).magnitude
        return self._viscosity_err * conv

    def set_viscosity_err(self, viscosity_err, unit='Pa*s'):
        """
        Set the viscosity.

        Arguments
        ---------
        viscosity_err : float or callable
            If callable, the function is used as
            absolute(viscosity_err(temp) * temp_err), whit temp in Kelvin.
            Otherwise the float value is used as error in the viscosity
            estimate.
        unit : str
            unit of the viscosity
        """
        T = self.get_temp(unit='K')
        dT = self.get_temp_err(unit='K')
        try:
            self._viscosity_err = absolute(viscosity_err(T) * dT)
        except:
            self._viscosity_err = viscosity_err
        self._viscosity_unit = str2u(unit)

    @property
    def density_medium(self):
        return self._density_medium

    def get_density_medium(self, unit=None):
        if unit is None or unit == self._density_medium_unit:
            conv = 1.0
        else:
            conv = ureg(self._density_medium_unit).to(unit).magnitude
        return self._density_medium * conv

    def set_density_medium(self, density, unit='Pa*s'):
        self._density_medium = density
        self._density_medium_unit = str2u(unit)

    def copy(self):
        return copy.copy(self)

    def get_dict(self):
        """
        Return data inside a dictionary.
        """
        d = OrderedDict()

        d[co.temp] = self.temp
        d[co.temp_err] = self.temp_err
        d[co.temp_unit] = self._temp_unit

        d[co.radius] = self.radius
        d[co.radius_err] = self.radius_err
        d[co.radius_unit] = self._radius_unit

        d[co.height] = self.height
        d[co.height_unit] = self._height_unit

        d[co.material] = self.material.name
        if self.material.name not in MATERIALS:
            d[co.density_p] = self.density_particle
            d[co.density_p_unit] = str2u('kg/m**3')

        d[co.medium] = self.medium
        if self.medium != 'water':
            d[co.density_m] = self.density_medium
            d[co.density_m_unit] = self._density_medium_unit

            d[co.viscosity] = self.viscosity
            d[co.viscosity_unit] = self._viscosity_unit

        return d


class PSDMeasurement(object):
    def __init__(self,
                 exp_setting=None,
                 freq_unit='Hz',
                 psd_unit='V**2/Hz',
                 warn=True):
        """
        Manage power spectra (PSDs) of one measurement.

        One-sided PSDs are assumed! They are stored in self.psds.
        Use **load()** to read data from a PSD data file or
        **add_psd()** to add **psd**-objects one after another.

        Arguments
        ---------
        exp_setting : ExpSetting
            Object discribing the experimental setting. Satndard values are
            assumed, if None. A warning will show up if warn is not False.
        freq_unit : str
            Unit of the frequency values ('Hz').
        psd_unit : str
            Unit fo the psd values ('V**2/Hz').
        warn : bool
            Show warnings if standard values are assumend.
        """
        if not exp_setting:
            temp = 25.0
            radius = 0.5
            exp_setting = ExpSetting(temp, radius,
                                     temp_unit='degC',
                                     radius_unit='um',
                                     warn=warn)
            if warn:
                warnings.warn('Standard values for experimental setting used! '
                              'I.e. temp=25 degC, radius=0.5 um, height=inf,'
                              'material=Polystyrene')

        if not isinstance(exp_setting, ExpSetting):
            raise Exception('exp_setting is not an ExpSetting object.')

        self.exp_setting = exp_setting

        self.psds = OrderedDict()

        self._freq_unit = freq_unit
        self._psd_unit = psd_unit

        self.active_calibration = False
        self.ex_axis = None
        self.ex_freq = None
        self.ex_amplitude = None
        self.ex_amplitude_err = None
        self._ex_amplitude_unit = None

        self.ex_power = None
        self.ex_power_err = None

        self._ex_power_unit = None

    def add_psd(self, name, psd):
        """
        Add a **psd** object to the measurement.

        Arguments
        ---------
        name : str
            String specifying the name of the axis.
        psd : PSD
            Object of the **PSD** class
        """
        # The first psd that's added set's the standard values of other psd
        # that might be added later on.
        if name in self.get_names():
            warnings.warn("PSD-object with name '{}' already "
                          "present. Data overridden".format(name))
        psd.name = name
        self.psds[name] = psd

    @property
    def names(self):
        """names of the axes."""
        return self.get_names()

    def get_names(self):
        """Return the names of the axes."""
        lst = [psd.name for psd in self.psds.values()]
        lst.sort()
        return lst

    def get_f_sample(self, name, unit=None):
        """
        Return fampling frequency of the PSD with the given name.
        """
        if unit is None:
            unit = self._freq_unit
        return self.psds[name].get_f_sample(unit=unit)

    def get_f_resolution(self, name, unit=None):
        if unit is None:
            unit = self._freq_unit
        unit_psd = self.psds[name]._freq_unit
        if unit_psd != unit:
            conv = ureg(unit_psd).to(unit).magnitude
        else:
            conv = 1.0
        return self.psds[name].df * conv

    def get_N_samples(self, name):
        return self.psds[name].N_samples

    def get_laterality(self):
        """
        Return a dictionary with names of the psds as keys and values
        discribing if the direction is lateral (True) or axial (False).
        """
        d = {name: psd.is_lateral()
             for name, psd in self.psds.items()}
        return d

    @property
    def freq_x(self):
        return self.get_freq('x')

    @property
    def freq_y(self):
        return self.get_freq('y')

    @property
    def freq_z(self):
        return self.get_freq('z')

    def get_freq(self, name, **kwargs):
        """
        Return the frequency vector of the axis specified by **name**.

        Arguments
        ---------
        name : str
            name of the axis.

        Keyword Arguments
        -----------------
        unit : str
        get_all : bool
            If False uses the internal mask.
        get_masked : bool
            If True returns the masked instead of the unmasked elements.
        """
        return self.psds[name].get_freq(**kwargs)

    @property
    def psd_x(self):
        return self.get_psd('x')

    @property
    def psd_y(self):
        return self.get_psd('y')

    @property
    def psd_z(self):
        return self.get_psd('z')

    def get_psd(self, name, **kwargs):
        """
        Return the psd vector of the axis specified by 'name'.

        Arguments
        ---------
        name : str
            name of the axis.
        kwargs
            can be:
             unit : str
             get_all : bool
                 If False uses the internal mask.
             get_masked : bool
                 If True returns the masked instead of the unmasked elements.
        """
        return self.psds[name].get_psd(**kwargs)

    @property
    def err_x(self):
        return self.get_psd_err('x')

    @property
    def err_y(self):
        return self.get_psd_err('y')

    @property
    def err_z(self):
        return self.get_psd_err('z')

    def get_psd_err(self, name, **kwargs):
        """
        Return the error vector of the axis specified by 'name'.

        Arguments
        ---------
        name : str
            name of the axis.
        kwargs can be:
        unit : str
        get_all : bool
            If False uses the internal mask.
        get_masked : bool
            If True returns the masked instead of the unmasked elements.
        offset : bool
            Whether to include the value at zero hertz.
        """
        return self.psds[name].get_err(**kwargs)

    def set_ac_params(self,
                      ex_axis,
                      ex_freq,
                      ex_amplitude,
                      ex_power,
                      ex_amplitude_err=0,
                      ex_power_err=0,
                      freq_unit='Hz',
                      amplitude_unit='m',
                      power_unit='V**2'):
        """
        Set the values of an active PSD measurement.

        Arguments
        ---------
        ex_axis : str
            Defines the name of the psd that was excited this must be one
            of self.names.
        ex_freq : float
            Frequency at which the bead was driven.
        ex_amplitude : float
            Amplitude of the sine movement of the bead w.r.t. the trap center.
        amplitude_unit : str
            Unit that the amplitude is given in. The amplitude is then
            converted to the same unit as the radius unit.
        ex_power : float
            The experimentally determined power at the given frequency.
        power_unit : str
            Unit the power is given in. This should be equal to
            psd unit * freq unit.
        ex_amplitude_err : float
            Error of the amplitude.
        ex_power_err : float
            Error of the measured power.

        Note
        ----
        It is good practise to measure the driving amplitude, e.g. via the
        piezo stage monitor signal, instead of taking the set amplitude for
        granted. The set value often differs from the actual one,
        especially at high driving frequencies.

        Reference
        ---------
        Tolić-Nørrelykke, et al. (2006)
            Part IV A
            Calibration of optical tweezers with positional detection in
            the back focal plane.
            Review of Scientific Instruments, 77(10), 103101.
            http://doi.org/10.1063/1.2356852
        """
        if ex_axis not in self.names:
            raise Exception('Unknown axis {}'.format(ex_axis))
        else:
            self.ex_axis = ex_axis

        if freq_unit != self._freq_unit:
            conv = ureg(freq_unit).to(self._freq_unit).magnitude
        else:
            conv = 1.0
        self.ex_freq = float(ex_freq) * conv

        r_unit = self.exp_setting._radius_unit
        if amplitude_unit != r_unit:
            conv = ureg(amplitude_unit).to(r_unit).magnitude
        else:
            conv = 1.0
        self.ex_amplitude = float(ex_amplitude) * conv
        self.ex_amplitude_err = float(ex_amplitude_err) * conv
        self._ex_amplitude_unit = r_unit

        self.ex_power = float(ex_power)
        self.ex_power_err = float(ex_power_err)
        self._ex_power_unit = power_unit

        self.active_calibration = True

    def get_stokes_drag(self, unit='N*s/m'):
        """
        Return Stoke's drag in the specified unit.

        The function uses the drag function of pyotc.physics evaluated at
        f=0 and height=inf.

        See Also
        --------
        pyotc.physics.drag
        """
        d = drag(self.exp_setting.get_radius(unit='m'),
                 self.exp_setting.get_temp(unit='K'),
                 density=self.exp_setting.get_density_medium(unit='kg/m**3'),
                 viscosity=self.exp_setting.get_viscosity(unit='Pa*s')).real
        if unit == 'N*s/m':
            conv = 1.0
        else:
            conv = ureg('N*s/m').to(unit).magnitude
        return d * conv

    def get_stokes_drag_err(self, unit='N*s/m'):
        """
        Return the error to Stoke's drag in the specified unit.

        The function calculates the error on the calculated stokes drag from
        the relative errors of the viscosity and the radius.

        See Also
        --------
        get_stokes_drag
        """
        v = self.exp_setting.get_viscosity()
        dv = self.exp_setting.get_viscosity_err()

        dr = self.exp_setting.get_radius_err(unit='m')
        r = self.exp_setting.get_radius(unit='m')
        ddrag = (dv / v + dr / r) * self.get_stokes_drag(unit=unit)
        return ddrag

    def get_corrected_drag(self,
                           mode=1,
                           distance=None,
                           focal_shift=1.0,
                           distance_unit='um',
                           drag_unit='N*s/m'):
        """
        Correct stokes drag for presence of either one wall or two parallel
        walls at the given height of the trapped particle.

        Only parallel movement of the sphere is considered!

        mode = 1
            Faxén's correction: drag = drag_stokes * faxen_factor
        mode = 2
            Linear superposition approach as done by Oseen in 1927; you need
            to provide the thickness of the sample!

            drag = drag_stokes * (faxen_factor(h) + faxen_factor(H-h)-1)

        Arguments
        ---------
        mode : 1 or 2
            If one or two walls are present.
        distance : float
            The distance between two parallel walls.
        focal_shift : float defaults to 1.0
            Relative shift of the focal point of the trap, due to diffraction.
            This usually is about 0.8 for an oil-immersion objective and a
            sample in water. The given height is automatically corrected for
            that shift.
        distance_unit : str
            Unit of distance.
        drag_unit : str
            unit of the output drag.
        """
        drag_stokes = self.get_stokes_drag(unit=drag_unit)
        radius = self.exp_setting.get_radius(unit=distance_unit)
        height = self.exp_setting.get_height(unit=distance_unit) * focal_shift

        if height < 1.5 * radius:
            warnings.warn("Faxén's correction does not work for heights lower"
                          "than 1.5 * radius.")

        if mode == 1:
            drag = drag_stokes * faxen_factor(height, radius)
        elif mode == 2:
            if distance is None:
                raise Exception('Distance between the two walls is not given.')
            if height > (distance - 1.5 * radius):
                warnings.warn('The correction does not work for heights above'
                              'thickness - 1.5 * radius.')
            drag = drag_stokes * oseen_factor(height, radius, distance)
        else:
            raise Exception('Unknown mode {}'.format(mode))
        return drag

    def save(self,
             directory=None,
             datafile=None,
             suffix='_psd_parameters.txt',
             datafile_extension='.dat',
             include_errors=False
             ):
        """
        Save PSD measurment to data and parameter file.

        Arguments
        ---------
        directory : path
            If None is given and there is no attribute 'directory' the current
            working directory is used.
        datafile : str
            If None, and there is no attribute 'datafilename' the current date
            and time is used.
        suffix : str
            Suffix for the parameter file.
        datafile_extension : str
            Extension of the data file.
        """
        # save data in data file
        if directory is None:
            try:
                directory = self.directory
            except:
                directory = './'
        if datafile is None:
            try:
                datafile = self.datafilename
            except:
                dstr = time.strftime("%Y-%m-%d_%H-%M")
                pfix = '_psd_measurement_'
                datafile = pfix + dstr

        if not datafile.endswith(datafile_extension):
            datafile += datafile_extension

        ptdfile = join(directory, datafile)

        # check that frequency vectors have the same lengths
        # TODO alternative: save freq vectors with axis name 'freq_x' etc.
        flengths = [len(self.get_freq(ax, get_all=True, offset=True))
                    for ax in self.names]
        if len(set(flengths)) > 1:
            raise Exception('Frequency vectors have different lengths!')

        psd_dict = OrderedDict()
        psd_dict.update({'psd_' + ax: self.get_psd(ax,
                                                   get_all=True,
                                                   offset=True)
                         for ax in self.names})
        if include_errors:
            psd_dict.update({'err_' + ax: self.get_psd_err(ax,
                                                           get_all=True,
                                                           offset=True)
                             for ax in self.names})

        plengths = [len(psd) for psd in psd_dict.values()]
        if len(set(plengths)) > 1:
            raise Exception('PSD vectors have different lengths!')

        freq = self.get_freq(self.names[0], get_all=True, offset=True)

        save_psd_data(ptdfile, freq, psd_dict)

        # save parameters in parameter file
        pfile = datafile[:datafile.rfind(datafile_extension)] + suffix
        ptpfile = join(directory, pfile)

        # generate parameters for parameter file
        params = OrderedDict()
        # names
        params[co.names] = ','.join(self.names)
        # N_avg
        N_avg_ = [self.psds[name].N_avg for name in self.names]
        params[co.N_avg] = ','.join(str(n) for n in N_avg_)
        # f_sample
        fs_ = [self.get_f_sample(name, unit=self._freq_unit)
               for name in self.names]
        params[co.f_sample] = ','.join(str(f) for f in fs_)

        params[co.freq_unit] = self._freq_unit
        params[co.psd_unit] = self._psd_unit

        params.update(self.exp_setting.get_dict())

        if self.active_calibration:
            ac_params = OrderedDict()
            ac_params[co.ex_axis] = self.ex_axis
            ac_params[co.ex_freq] = self.ex_freq
            ac_params[co.ex_amp] = self.ex_amplitude
            ac_params[co.ex_amp_err] = self.ex_amplitude_err
            ac_params[co.ex_pow] = self.ex_power
            ac_params[co.ex_pow_err] = self.ex_power_err
            ac_params[co.ex_amp_unit] = self.exp_setting._radius_unit
            ac_params[co.ex_pow_unit] = self._ex_power_unit
        else:
            ac_params = None

        save_psd_params(ptpfile, params, ac_param_dict=ac_params)

        self.directory = directory
        self.datafilename = datafile
        self.paramfile = pfile

    def load(self,
             directory,
             datafile,
             paramfile=None,
             suffix='_psd_parameters.txt'
             ):
        """
        Load the psd-data and according parrameter file.

        Arguments
        ---------
        directory : path
            Path to the data folder.
        datafile : str
            Filename of the psd-data file '*.dat'.
        paramfile : str
            Filename of the corresponding parameter file. If None, the
            extensiotn of the datafile is cut off and **suffix** is appended
        suffix : str
            Suffix to the datafile name that characterizes the parameter file.

        Note
        ----
        datafile
            The data file is assumed to have the data in columns, with their
            names in the first row. Normally the columns are called 'freq',
            'psd_x', psd_y, etc. Columns starting with 'psd_' are still loaded,
            and the strings after the underscore '_' are taken as names for the
            created **psd** objects. So 'PSD_x1', 'PSD_x2', etc. will also
            work and create axes with names x1, x2, etc.
            **Important:** Any name that contains 'z' is considered to be an
            axis in the axial direction. Hence **psd.direction** is set to
            'axial'.
        paramfile
            The parameter file is read by the module **configparser**,
            so a **config-file standard** is assumed.
            Parameters of the [DEFAULT] section are read.
            The following parameters are read, the values in brackets are the
            fallback values - if none is given the value is mendatory:
             - freq_unit ('Hz')
             - n_avg
             - sampling_rate or f_sample
             - psd_unit ('V**2/Hz')
             - bead_dia or radius
             - err_bead_dia or radius_err (0.0)
             - diameter_unit or radius_unit ('um')
             - density and density_unit or material ('PS')
             - density_med (None - water is used later on)
             - density_med_unit ('kg/m**3')
             - viscosity (None - water is used later on)
             - viscosity_unit ('Pa*s')
             - height (inf)
             - height_unit ('um')
             - temperature (25)
             - temperature_error (5)
             - temp_unit ('celsius')
            If the trapped particle was actively driven by a sine-wave and
            an active calibration should be done, the following values must be
            in the 'ACTIVE_CALIBRATION' section of the parameter file.
             - excitation_axis
             - excitation_frequency
             - excitation_amplitude
             - excitation_amplitude_error (0)
             - amplitude_unit (um)
             - reference power or power
             - reference_power_error or power_err
             - power_unit ('V**2')
        """
        dfile = join(directory, datafile)

        if paramfile is None:
            paramfile = datafile[:-4] + suffix

        self.paramfile = paramfile

        pfile = join(directory, paramfile)

        # read psd data file
        data = read_std_data_file(dfile, lower_names=True)

        # read parameter file
        pars = read_PSD_parameter_file(pfile)

        # set up fallback values
        params = {co.freq_unit: 'Hz',
                  co.psd_unit: 'V**2/Hz',
                  co.radius_err: 0.0,
                  co.radius_unit: 'm',
                  co.height: inf,
                  co.height_unit: 'm',
                  co.temp: 25.0,
                  co.temp_err: 2.5,
                  co.temp_unit: 'celsius',
                  co.material: '',
                  co.density_p: None,
                  co.density_p_unit: 'kg/m**3',
                  co.medium: 'water',
                  co.density_m: None,
                  co.density_m_unit: 'kg/m**3',
                  co.viscosity: None,
                  co.viscosity_unit: 'Pa*s'}
        # overwrite default parameters
        params.update(pars)

        # backward compatibility to version 0.2.2
        if co.names not in params:
            # get the axis names from the header in the datafile
            names = [name.split('_')[1]
                     for name in data.keys() if name.lower().startswith('psd')]
        else:
            names = params[co.names].split(',')

        # check if N_avg and f_sample are specified
        for par in [co.N_avg, co.f_sample]:
            if par not in params:
                raise Exception('Parameter {} missing in parameter file {}'
                                ''.format(par, pfile))

        n_avg_ = params[co.N_avg].split(',')
        # backward compatibility to version 0.2.2
        if len(n_avg_) == 1:
            n_avg_ = [n_avg_[0] for name in names]
        N_avg = {}
        for name, n_avg in zip(names, n_avg_):
            N_avg[name] = int(n_avg)

        f_sample_ = params[co.f_sample].split(',')
        # backward compatibility to version 0.2.2
        if len(f_sample_) == 1:
            f_sample_ = [f_sample_[0] for name in names]
        f_sample = {}
        for name, fs in zip(names, f_sample_):
            f_sample[name] = float(fs)

        # check if radius or diameter was specified
        if co.radius not in params:
            raise Exception('Bead radius is not specified in '
                            'parameter file {}'.format(paramfile))

        expset = ExpSetting(params[co.temp],
                            params[co.radius],
                            temp_err=params[co.temp_err],
                            radius_err=params[co.radius_err],
                            height=params[co.height],
                            density_particle=params[co.density_p],
                            density_medium=params[co.density_m],
                            viscosity=params[co.viscosity],
                            material=params[co.material],
                            medium=params[co.medium],
                            temp_unit=params[co.temp_unit],
                            radius_unit=params[co.radius_unit],
                            height_unit=params[co.height_unit],
                            density_particle_unit=params[co.density_p_unit],
                            density_medium_unit=params[co.density_m_unit],
                            viscosity_unit=params[co.viscosity_unit])
        self.exp_setting = expset

        # now get teh freq vector and psd data and put it into a PSD object
        freq = data.pop(co.freq)

        for name in names:
            try:
                psd_vals = data['psd_' + name]
            except:
                warnings.warn('PSD value for axis name {} not found'
                              ''.format(name))
                continue
            try:
                psd_err = data['err_' + name]
            except:
                psd_err = None

            psd = PSD(freq,
                      psd_vals,
                      err=psd_err,
                      name=name,
                      f_sample=f_sample[name],
                      N_avg=N_avg[name],
                      freq_unit=params[co.freq_unit],
                      psd_unit=params[co.psd_unit]
                      )
            self.add_psd(name, psd)

        # check if there's active calibration information
        if params['active_calibration']:
            ex_axis = params[co.ex_axis]
            ex_freq = params[co.ex_freq]
            ex_amplitude = params[co.ex_amp]
            ex_power = params[co.ex_pow]

            if co.ex_amp_err in params:
                ex_amplitude_err = params[co.ex_amp_err]
            else:
                ex_amplitude_err = 0

            if co.ex_pow_err in params:
                ex_power_err = params[co.ex_pow_err]
            else:
                ex_power_err = 0

            if co.ex_amp_unit in params:
                ex_amp_unit = params[co.ex_amp_unit]
            else:
                ex_amp_unit = self._radius_unit

            if co.ex_pow_unit in params:
                power_unit = params[co.ex_pow_unit]
            else:
                psd_unit = params[co.psd_unit]
                freq_unit = params[co.freq_unit]
                power_unit = str((ureg(psd_unit) * ureg(freq_unit)).units)

            self.set_ac_params(ex_axis,
                               ex_freq,
                               ex_amplitude,
                               ex_power,
                               ex_amplitude_err=ex_amplitude_err,
                               ex_power_err=ex_power_err,
                               freq_unit=params[co.freq_unit],
                               amplitude_unit=ex_amp_unit,
                               power_unit=power_unit)

        self.directory = directory
        self.datafilename = datafile

    def get_ex_freq(self, unit='Hz'):
        if unit != self._freq_unit:
            conv = ureg(self._freq_unit).to(unit).magnitude
        else:
            conv = 1.0
        return self.ex_freq * conv

    def get_ex_amplitude(self, unit=None):
        if unit is None or unit == self._ex_amplitude_unit:
            return self.ex_amplitude
        else:
            conv = ureg(self._ex_amplitude_unit).to(unit).magnitude
            return self.ex_amplitude * conv

    def get_ex_amplitude_err(self, unit=None):
        if unit is None or unit == self._ex_amplitude_unit:
            return self.ex_amplitude_err
        else:
            conv = ureg(self._ex_amplitude_unit).to(unit).magnitude
            return self.ex_amplitude_err * conv

    def get_ex_power(self, unit=None):
        if unit is None:
            conv = 1.0
        else:
            conv = ureg(self._ex_power_unit).to(unit).magnitude
        return self.ex_power * conv

    def get_ex_power_err(self, unit=None):
        if unit is None:
            conv = 1.0
        else:
            conv = ureg(self._ex_power_unit).to(unit).magnitude
        return self.ex_power_err * conv

    def reset_masks(self):
        for psd in self.psds.values():
            psd.reset_mask()

    def exclude_freq(self, f_ex, names=None):
        """
        Exclude data points at frequencies f_ex.

        Arguments
        ---------
        f_ex : float or list(floats) or None
            Frequencies to be excluded. If None, the data point at the
            excitation frequency of the excited axis is excluded.
        name : str
            Name of the psd where the data point shall be excluded. If None,
            all psds get f_ex excluded
        """
        if names and not isinstance(names, list):
            names = [names]

        for name in names if names else self.names:
            self.psds[name].exclude_freq(f_ex)

    def exclude_freq_outside(self, fmin, fmax, names=None, reset_mask=False):
        """
        Exclude values outside the range fmin, fmax.
        """
        if names and not isinstance(names, list):
            names = [names]

        for name in names if names else self.names:
            if reset_mask:
                self.psds[name].reset_mask()
            self.psds[name].exclude_freq_outside(fmin, fmax)

    def plot_psds(self, names=None, axis=None, **kwargs):
        """
        Plots all power spectral densities or the specified name only.

        Calls the function plot_psd of the PSD object.

        Arguments
        ---------
        name : str or None
        axis : matplotlib.Axis or None
            Axis to add the plot to.
        kwargs : keyword arguments passed to PSD.plot_psd()
            e.g.:
             - plot_all=False
             - plot_masked=False
             - plot_errors=False
             - plot() keyword argumets

        Returns the figure object.
        """
        if names and not isinstance(names, list):
            names = [names]

        if 'title' not in kwargs.keys():
            kwargs['title'] = ('PSDs at {0:1.3f} µm'.format(self.exp_setting.get_height(unit='um')))

        for name in names if names else self.names:
            fig = self.psds[name].plot_psd(axis=axis, **kwargs)
            if axis is None:
                axis = fig.axes[0]

        return fig


def gen_psdm_from_region(region, T_msr, N_avg,
                         T_delay=0.0, psd_traces=None,
                         active_calibration=False, position_traces=None,
                         ex_freq=None,  position_unit='um', exp_setting=None):
    """
    """
    pm = PSDMeasurement(exp_setting=exp_setting, warn=True)

    psd_traces = psd_traces or region.traces
    N_min = int(T_delay * region.samplingrate)
    N_max = int(T_msr * region.samplingrate)
    samples = slice(N_min, N_min + N_max)
    data = region.get_data(psd_traces, samples=samples)

    for name, dat in zip(psd_traces, data.T):
        p = gen_PSD_from_time_series(dat, region.samplingrate, N_avg)
        pm.add_psd(name, p)

    if active_calibration:
        if not ex_freq:
            raise Exception('Excitation frequency must be provided for an'
                            ' active calibration.')

        position_traces = position_traces or ['positionX', 'positionY']
        ex_axis = region._excited(position_traces)
        stage_ex_axis = position_traces[ex_axis]
        stage_signal = region.get_data(stage_ex_axis, samples=samples)[:, 0]
        p = gen_PSD_from_time_series(stage_signal,
                                     region.samplingrate,
                                     N_avg,
                                     calc_errors=True)

        ## !!! important this is already in µm, due to the setup-specific
        ## config file
        ex_amp = float(sqrt(2 * p.psd[p.freq == ex_freq] * p.df))
        ex_amp_err = float(sqrt(2 * p.psd_err[p.freq == ex_freq] * p.df))

        psd_ex_axis = psd_traces[ex_axis]

        freqs = pm.get_freq(psd_ex_axis)
        ex_pow = float(pm.get_psd(psd_ex_axis)[freqs == ex_freq] *
                       pm.psds[psd_ex_axis].df)
        ex_pow_err = float(pm.get_psd_err(psd_ex_axis)[freqs == ex_freq] *
                           pm.psds[psd_ex_axis].df)

        pm.set_ac_params(psd_ex_axis, ex_freq, ex_amp, ex_pow,
                         ex_amplitude_err=ex_amp_err, ex_power_err=ex_pow_err,
                         amplitude_unit=position_unit)

    return pm
#####--------------------------------------------------------------------------
#---- shape ---
#####--------------------------------------------------------------------------

def lorentzian_psd(freq, D, f_c):
    """
    Return the values for frequenzies f of a Lorentzian-shaped **one-sided**
    power spectral density with a diffusion constant D and a corner frequency
    f_c.

    The function applied is: psd = D / (pi**2 * (freq**2 + f_c**2)).

    Arguments
    ---------
    freq : array(float)
        Frequency vector.
    D : float
        Diffusion constant.
    f_c : float
        Corner frequency.

    Returns
    -------
    array

    References
    ----------
    Equ. (9) in:
        Tolić-Nørrelykke et al. (2006)
        Calibration of optical tweezers with positional detection in the back
        focal plane. Review of Scientific Instruments, 77(10), 103101.
        http://doi.org/10.1063/1.2356852
    """
    l = D / (pi**2 * (freq**2 + f_c**2))
    return l


def hydro_psd(freq,  # Hz
              D,  # arb**2/s
              f_c,  # Hz
              radius=0.5e-6,  # m
              height=inf,  # m
              temp=293.15,  # K
              rho=1000,  # kg/m**3
              density_med=None,  # kg/m**3
              viscosity=None,  # Pa*s
              lateral=True,
              verbose=False):
    """
    The hydrodynamically correct power spectral density of a sphere at a given
    height in a viscous medium.

    The function takes Faxén's law and the mass of the sphere into account.

    Arguments
    ---------
    freq : array(float)
        Frequency vector in Hertz
    D : float
        Diffusion coefficient (with no correction, i.e. pure Stokes drag).
    f_c : float
        Corner frequency in Hz.
    radius : float
        Radius of the sphere in meters.
    height : float
        Height, i.e. the bead-center -- surface distance (the real one) in
        meters.
    temp : float
        Absolute temperature in Kelvin
    rho : float
        Mass density of the sphere in kg/m³
    density_med : float
        Mass density of the medium in kg/m³. If None, the density of water at
        the given temperatrue is used.
    viscosity : float
        Viscosity of the medium in Ns/m². If None, the viscosity of water at
        temperature temp is assumed.
    lateral : bool
        deprecated.
    verbose : bool
        be verbose.

    References
    ----------
    [1] Appendix D in:
        Tolić-Nørrelykke et al. (2006)
        Calibration of optical tweezers with positional detection in the back
        focal plane. Review of Scientific Instruments, 77(10), 103101.
    [2] Berg-Sørensen, K., & Flyvbjerg, H. (2005). The colour of thermal noise
        in classical Brownian motion: A feasibility study of direct
        experimental observation. New Journal of Physics, 7.

    See Also
    --------
    viscosity_H2O

    """
    drag_stokes = drag(radius, temp, density=density_med,
                       viscosity=viscosity)
    f_c0 = f_c
    drag_l = drag(radius, temp, freq=freq, height=height, density=density_med,
                  viscosity=viscosity, lateral=lateral, verbose=verbose)
    rel_drag = drag_l / drag_stokes

    # see Ref. [2]
    # f_m0 = drag(radius) / (2 * pi * m*)
    # m* = m_p + 2/3*pi*R³*rho_fluid
    m_p = 4/3 * pi * radius**3 * rho
    f_m0 = drag_stokes / (2 * pi * (m_p + 2/3 * pi * radius**3 * density_med))

    P = (D * rel_drag.real /
         (pi**2 * ((f_c0 + freq * rel_drag.imag - freq**2 / f_m0)**2 +
                   (freq * rel_drag.real)**2)))

    if verbose:
        print('hydro_psd:')
        print('Stokes drag (Ns/m) = {:1.4e}'.format(drag_stokes))
        print('f_c (Hz) = {:1.4e}'.format(f_c0))
        print('corrected drag (Faxen) (Ns/m)  = {:1.4e}'.format(drag_l))
        print('relative drage  = {:1.4e}'.format(rel_drag))
        print('mass sphere (kg) = {:1.4e}'.format(m_p))
        print('f_m0 (Hz) = {:1.4e}'.format(f_m0))
        print('P (arb²/Hz)  = {:1.4e}'.format(P))
        
    return P.real


def low_pass_filter(freq, f3dB, alpha=0):
    """
    Produces the relative PSD of a signal that, to a factor (1-alpha²), is
    filtered by a 1st-order low-pass filter.

    This, in particular, describes the filtering of a quadrant photo-diode:
    Only a fraction of the photons that reach the diode is absorped in the
    depletion region, the other fraction get's absorped outside that region.
    The electron-hole pairs need to diffuse to the depletion region until
    they produce a photo-current. This process causes an effective low-pass
    filter for these photons. The in the power spectrum the function looks
    like this:

     F(f) = alpha² + (1 - alpha²) / (1 + (f / f3dB)**2)

    Arguments
    ---------
    freq : array(float)
        Frequency vector.
    f3dB : float
        Cut-off frequency of the low-pass filter.
    alpha : float
        Filter efficiency. Only a fraction (0 <= alpha <= 1) of the signal that
        is not low pass filtered. Thus, alpha = 1, will produce no low-pass
        filter at all, whereas alpha = 0, will produce a 1st-order low-pass
        filter.

    References
    ----------
    Equ. (35) in:
        Berg-Sørensen, K., & Flyvbjerg, H. (2004)
        Power spectrum analysis for optical tweezers. Review of Scientific
        Instruments, 75(3), 594–612.
        http://doi.org/10.1063/1.1645654
    Equ. (20) in:
        Berg-So̸rensen, K., et al. (2006)
        Power spectrum analysis for optical tweezers. II: Laser wavelength
        dependence of parasitic filtering, and how to achieve high bandwidth.
        Review of Scientific Instruments, 77(6), 063106.
        http://doi.org/10.1063/1.2204589
    """
    return alpha**2 + (1 - alpha**2) / (1 + (freq / f3dB)**2)


def apply_low_pass_filter(fun, f3dB, alpha):
    """
    Decorator function to produces a low-pass filtered modification of the
    input PSD function

    See also
    --------
    low_pass_filter()

    Arguments
    ---------
    fun : function
        Function with first positional argument being the frequency vector
        'freq'.
    f3dB : float
        Cut-off frequency of the low-pass filter.
    alpha : float
        Filter efficiency. Only a fraction (0 <= alpha <= 1) of the signal that
        is not low pass filtered. Thus, alpha = 1, will produce no low-pass
        filter at all, whereas alpha = 0, will produce a 1st-order low-pass
        filter.

    Returns
    -------
    function with arguments (freq, *args, **kwargs).

    References
    ----------
    Equ. (35) in:
        Berg-Sørensen, K., & Flyvbjerg, H. (2004)
        Power spectrum analysis for optical tweezers. Review of Scientific
        Instruments, 75(3), 594–612.
        http://doi.org/10.1063/1.1645654
    Equ. (20) in:
        Berg-So̸rensen, K., et al. (2006)
        Power spectrum analysis for optical tweezers. II: Laser wavelength
        dependence of parasitic filtering, and how to achieve high bandwidth.
        Review of Scientific Instruments, 77(6), 063106.
        http://doi.org/10.1063/1.2204589
    """
    def lp_filtering(freq, *args, **kwargs):
        f = (low_pass_filter(freq, f3dB, alpha=alpha) *
             fun(freq, *args, **kwargs))
        return f

    if hasattr(fun, '__name__'):
        lp_filtering.__name__ = ('low-pass filtered "{0:s}"'
                                 ''.format(fun.__name__))
    if hasattr(fun, '__doc__'):
        lp_filtering.__doc__ = ('Low pass filtered function with cut-off '
                                'frequency f3dB = {0:1.1f} Hz and efficiency '
                                'alpha = {1:1.3f} \n\n'
                                'calls\n{2}{3} \n\n'
                                'Documentation of {2}:\n{4}'
                                ''.format(f3dB, alpha, fun.__name__,
                                          str(signature(fun)), fun.__doc__))
    return lp_filtering


def apply_aliasing(fun, f_sample, N=9):
    """
    Decorator function to produce an aliased version of the input PSD function.

    The aliased version of a psd is calculated by adding the psd values of
    frequencies that lie beyound the sampling frequency to the range below the
    sampling frequency.

     PSD_aliased(f) = sum(PSD(f + n * f_sample)) from n=-N to N.

    Note
    ----
    Note, that N is actually infinity, but a fintite number of summations
    should be sufficient to account for aliasing. N=5 gives about 1.5% error at
    high frequencies, N=10 should give less than 0.5%.

    Arguments
    ---------
    fun : function
        Function to be aliased.
    f_sample : float
        Sampling frequency
    N : int
        Number that defines how many ranges of f_sample should be taken into
        account. The default N=9 give a very good approximation with deviations
        of less than 0.5%
    *args
        passed to the given function.
    **kwargs
        passed to the given function.

    Returns
    -------
    function with arguments (freq, *args, **kwargs)

    References
    ----------
    Equ. (37) in:
        Berg-Sørensen, K., & Flyvbjerg, H. (2004)
        Power spectrum analysis for optical tweezers. Review of Scientific
        Instruments, 75(3), 594–612.
        http://doi.org/10.1063/1.1645654
    """
    def aliasing(freq, *args, **kwargs):
        aliased = sum(fun(freq + i * f_sample, *args, **kwargs)
                      for i in range(-N, N+1))
        return aliased

    if hasattr(fun, '__name__'):
        aliasing.__name__ = 'aliased "{0:s}"'.format(fun.__name__)
    if hasattr(fun, '__doc'):
        aliasing.__doc__ = ('Aliased function with\n'
                            'N = {0} and f_sample = {1:1.1f} Hz. \n\n'
                            'Calls\n'
                            '{2}{3}\n\n'
                            'Documentation of {2}:\n{4}'
                            ''.format(N, f_sample, fun.__name__,
                                      str(signature(fun)), fun.__doc__))
    return aliasing



# -*- coding: utf-8 -*-
"""
Created on Wed Mar  9 17:03:58 2016

@author: Tobias Jachowski
"""
import persistent
from collections import deque


class Node(persistent.Persistent):
    """
    State information via references:

    Node implements a directed acyclic graph with depth and breadth first
    traversal. Each node has an attribute `cargo`, which can hold any python
    object. Node provides methods to get and set nodes (and cargo) based on
    different conditions, like level of depth or type of instance.
    """

    def __init__(self, cargo=None, **kwargs):
        """
        Attributes
        ----------
        cargo : object
            The cargo of the Node.
        """
        # initialize variables
        self._relatives = {'children': [],
                           'parents': []}
        self._max_relatives = {'children': -1,
                               'parents': -1}
        self.cargo = cargo
        # self._processed = None
        super().__init__(**kwargs)

    def add_child(self, child, index=None, after=None, before=None):
        """
        Add a child. See `add_relative()` for details.
        """
        return self.add_relative(child, child=True, index=index, after=after,
                                 before=before)

    def add_parent(self, parent, index=None, after=None, before=None):
        """
        Add a parent. See `add_relative()` for details.
        """
        return self.add_relative(parent, child=False, index=index, after=after,
                                 before=before)

    def add_relative(self, relative, child=True, index=None, after=None,
                     before=None, bidirectional=True):
        """
        Add a `relative` as a child or a parent.

        The index of the position of inserting the relative is determined by
        the following parameters in descending superseding order: `index`,
        `after`, and `before`.

        Parameters
        ----------
        relative : Node or cargo
            Node to be added as a relative, or cargo with attribute `_node`
            referencing a Node.
        child : bool, optional
            Add a child or a parent.
        index : int, optional
        after : Node or cargo, optional
        before : Node or cargo, optional
        bidirectional : bool, optional
            Add alsoe `self` to the `relative`, after `relative` has been added
            to `self`, i.e. not only reference the `relative` from `self`, but
            also reference `self` from `relative`.

        Returns
        -------
        bool
            True if `relative` could be added or `relative` is already present.
            Otherwise return False.
        """
        # Get nodes of relative, after, and before
        relative = _get_node(relative)
        if after is not None:
            after = _get_node(after)
        if before is not None:
            before = _get_node(before)

        # Try to add relative
        self_add = self._add_relative(relative, child=child, index=index,
                                      after=after, before=before)

        # Try to add self to relative
        relation_add = True
        if bidirectional:
            relation_add = relative._add_relative(self, child=not child)

        # check for circular reference
        circular_reference = relative.circular_reference(descendants=child)

        # Relative could not be added, or self could not be added to relative,
        # or circular_reference detected
        if not self_add or not relation_add or circular_reference:
            # Either relative couldn't be added or
            # circular reference occured -> remove relative
            self._remove_relative(relative, child=child)
            relative._remove_relative(self, child=not child)
            return False

        # relative was added or already present
        return True

    def _add_relative(self, relative, child=True, index=None, after=None,
                      before=None):
        if child:
            relatives = self._children
            max_relatives = self.max_children
        else:
            relatives = self._parents
            max_relatives = self.max_parents

        if max_relatives != -1 and len(relatives) >= max_relatives:
            print("Can not add relative. No more relatives allowed!")
            return False

        if relative is not None:
            if relative in relatives:
                # parent already exists, avoid adding a second time and return
                # True
                print("Relative %s already referenced." % relative)
                return True

            # determine the index of self._parents/self._childs before which
            # the parent/child should be inserted
            if index is not None:
                # highest priority of index determination, take the given index
                pass
            elif after in relatives:
                # second highest priority (determine index by after)
                index = relatives.index(after) + 1
            elif before in relatives:
                # third highest priority (determine index by before)
                index = relatives.index(before)
            else:
                # default: append parent/child (add parent/child at the end)
                index = len(relatives)

            # Add parent/child at determined index
            relatives.insert(index, relative)
            # inform ZODB of change
            self._p_changed = True

            return True

    def remove_child(self, child):
        return self.remove_relative(child, child=True)

    def remove_parent(self, parent):
        return self.remove_relative(parent, child=False)

    def remove_relative(self, relative, child=True):
        """
        Remove child or parent from self.
        """
        relative = _get_node(relative)

        self_remove = self._remove_relative(relative, child)
        relate_remove = relative._remove_relative(self, child=not child)

        # No relative was removed
        if not (self_remove or relate_remove):
            return False

        # relative was removed
        return True

    def _remove_relative(self, relative, child=True):
        if child:
            relatives = self._children
        else:
            relatives = self._parents

        # remove child/parent from relatives
        if relative in relatives:
            relatives.remove(relative)
            # inform ZODB of change
            self._p_changed = True
            return True

        return False

    def circular_reference(self, descendants=True):
        """
        Check for circular references.
        """
        # determine direction of search for circular reference
        relatives = self.relatives(descendants=descendants)

        # Search for a circular reference.
        # First, assume there is no circular reference:
        # If this View has no modifications or relatives, there is no
        # circular reference and circular stays False
        # Check for circular reference via modifications
        # for modification in modifications:
        #     circular = circular or modification.circular_reference(down,
        #                                                            caller)
        # Check for circular reference via relatives
        for relative in relatives:
            if relative is self:
                print("Circular reference detected -> Illegal connection!")
                return True

        # stop recursion: return result of circular recursive search
        return False

    def relatives(self, descendants=True, includeself=False, dft=True,
                  level=-1, cargo=False):
        """
        Traverse relatives.

        Parameters
        ----------
        descendants : bool
            Yield descendants or ancestors
        includeself : bool
            Yield only relatives or also include self
        dft : bool
            Use depth first or breadth first traversal
        level : int
            Yield self (0), up to first generation (1), up to second generation
            (2), ...
        cargo : bool
            Yield node or its cargo

        Yields
        -------
        Node
            If `cargo` is False.
        object
            If `cargo` is True.

        Notes
        -----
        Level describes the generation up to which a relative should be
        yielded:
                         |
        level 0         Node
                         |  \
        level 1          |   Node
                         |    |
        level 2          |   Node
                         |  / |  \
        level 1/3       Node Node Node
                       / |  \
        level 2/4  Node Node Node
                              |
        """
        # see http://jeremykun.com/2013/01/22/depth-and-breadth-first-search/

        # Initialization ...
        toprocess = deque()

        if dft:
            # reversed iteration for depth first search
            iterswitch = reversed
            addtoprocess = toprocess.append
        else:
            iterswitch = iter
            addtoprocess = toprocess.appendleft

        # start with either self or the relatives
        if includeself:
            # add self at level 0
            toprocess.append((0, self))
        elif level == -1 or level >= 1:
            # relatives are either children or parents:
            if descendants:
                relatives = self._children
            else:
                relatives = self._parents

            for n in iterswitch(relatives):
                # add relatives at level 1
                addtoprocess((1, n))

        visited = set()
        # better execution time but not thread save!
        # search_time = datetime.now().timestamp()
        # process_id = uuid.uuid4()
        # search_time = process_id
        # better execution time but not thread save!

        # Traversal and processing ...
        while len(toprocess) > 0:
            current_level, node = toprocess.pop()

            # if node._processed != search_time:
            if node not in visited:
                visited.add(node)
                # node._processed = search_time

                if cargo and node.cargo is not None:
                    yield node.cargo
                if not cargo:
                    yield node

                if descendants:
                    relatives = node._children
                else:
                    relatives = node._parents

                next_level = current_level + 1
                if level == -1 or next_level <= level:

                    for n in iterswitch(relatives):
                        # if n._processed != search_time:
                        if n not in visited:
                            addtoprocess((next_level, n))

    @property
    def _children(self):
        return self._relatives['children']

    @property
    def _parents(self):
        return self._relatives['parents']

    @property
    def max_children(self):
        return self._max_relatives['children']

    @max_children.setter
    def max_children(self, max_children):
        self._max_relatives['children'] = max_children

    @property
    def max_parents(self):
        return self._max_relatives['parents']

    @max_parents.setter
    def max_parents(self, max_parents):
        self._max_relatives['parents'] = max_parents

    @property
    def parents(self):
        return self.relatives(descendants=False, dft=False, level=1)

    @property
    def children(self):
        return self.relatives(descendants=True, dft=False, level=1)

    @property
    def ancestors(self):
        return self.relatives(descendants=False, dft=False)

    @property
    def descendants(self):
        return self.relatives(descendants=True, dft=False)


class GraphMember(persistent.Persistent):
    """
    GraphMember has one instance of class Node, offering convenience functions
    to handle the Node simply by inheriting, and give the ancestor class a Node
    like behaviour.
    It implements methods to inform other members of the graph of a change.
    """
    def __init__(self, max_parents=-1, max_children=-1, updated=False,
                 name=None, group=None, **kwargs):
        self._node = Node(cargo=self)

        self.max_parents = max_parents
        self.max_children = max_children

        # Initialize the status of a GraphMember to be not updated, per default
        self.updated = updated

        self.name = name
        self.group = group

    def members(self, name=None, group=None, instance_class=None,
                descendants=True, includeself=True, dft=False, level=-1):
        for relative in self._node.relatives(descendants=descendants,
                                             includeself=includeself, dft=dft,
                                             level=level, cargo=True):
            if (name is None or relative.name == name) \
                    and (group is None or relative.group == group) \
                    and (instance_class is None
                         or isinstance(relative, instance_class)):
                yield relative

    def group_ends(self, group, root=None, descendants=True, includeself=True):
        for member in self.members(group=group, descendants=descendants,
                                   includeself=includeself, dft=True):
            if member.is_end(root=root):
                yield member

    def group_roots(self, group, descendants=True, includeself=True):
        return self.group_ends(group, root=True, descendants=descendants,
                               includeself=includeself)

    def group_leafs(self, group, descendants=True, includeself=True):
        return self.group_ends(group, root=False, descendants=descendants,
                               includeself=includeself)

    def group_end(self, group, root=True, descendants=True, includeself=True):
        ends = self.group_ends(group, root=root, descendants=descendants,
                               includeself=includeself)
        return next(ends, None)

    def group_root(self, group, descendants=True, includeself=True):
        return self.group_end(group, root=True, descendants=descendants,
                              includeself=includeself)

    def group_leaf(self, group, descendants=True, includeself=True):
        return self.group_end(group, root=False, descendants=descendants,
                              includeself=includeself)

    def is_end(self, root=None):
        """
        Parameters
        ----------
        root : None or bool, optional
            If root is None, check whether GraphMember is root or leaf of
            own group. If root is True, check whether Graphmember is root of
            group, if root is False check Graphmember beeing a leaf of group.
        """
        if root is None:
            return self.is_group_root or self.is_group_leaf
        elif root:
            return self.is_group_root
        else:
            return self.is_group_leaf

    @property
    def is_group_root(self):
        parent = self.parent
        return parent is None or self.group != parent.group

    @property
    def is_group_leaf(self):
        child = self.child
        return child is None or self.group != child.group

    def set_changed(self, descendants=True, includeself=True, level=-1,
                    **kwargs):
        """
        Inform all descendants/ancestors (Nodes that have this Node as a
        parent/child) about a change, so that the children inform their
        children, about the change.

        Has to be called upon any change of `self` and parents.

        It calls `member_changed(ancestor=descendants, **kwargs)` on `self`,
        if `includeself` is True, and on all descendants.

        Parameters
        ----------
        descendants : bool
            Inform descendants or ancestors about change
        level : int
            Up to which generation the change should be proclaimed
        includeself : bool
            Should self be informed, too, i.e. should member_changed() be
            called?
        **kwargs
            See member_changed() for parameters
        """
        # inform node about change
        # the node in turn will call member_changed() method of the cargos
        # requested (see set_changed() method of node)
        if includeself:
            self.member_changed(ancestor=descendants, calledfromself=True,
                                **kwargs)

        # Get either descendants or ancestors to be informed of the change
        members = self.members(descendants=descendants, includeself=False,
                               dft=False, level=level)

        for member in members:
            member.member_changed(ancestor=descendants, calledfromself=False,
                                  **kwargs)

    def member_changed(self, ancestor=True, calledfromself=False,
                       updated=False, **kwargs):
        # `self` triggered the change. Set updated status of `self` according
        # to the parameter `updated`.
        if calledfromself:
            self.updated = updated
        # An ancestor triggered the change and `self` is set to be outdated. A
        # change of descendants will be ignored.
        if not calledfromself and ancestor:
            self.updated = False

    def add_member(self, member, child=True, index=None, after=None,
                   before=None, set_changed=True, **kwargs):
        # Try to add the new member
        added = self._node.add_relative(member, child=child, index=index,
                                        after=after, before=before, **kwargs)
        # Inform the child or self and the children about change
        if added and set_changed:
            if child:
                # Inform the child about the addition of a new parent (i.e. the
                # addition of self)
                member.set_changed()
            else:
                # inform self and children about addition of a new parent
                self.set_changed()
        return added

    def add_parent(self, parent, index=None, after=None, before=None,
                   set_changed=True, **kwargs):
        # Try to add the new parent
        return self.add_member(parent, child=False, index=index, after=after,
                               before=before, set_changed=set_changed,
                               **kwargs)

    def add_child(self, child, index=None, after=None, before=None,
                  set_changed=True, **kwargs):
        # Try to add the new child
        return self.add_member(child, child=True, index=index, after=after,
                               before=before, set_changed=set_changed,
                               **kwargs)

    def remove_member(self, member, child=True, set_changed=True):
        # Try to remove the member
        removed = self._node.remove_relative(member, child=child)

        # Inform the child or self and the children about change
        if removed and set_changed:
            if child:
                # Inform the child about the loss of a parent (i.e. the loss
                # of self)
                member.set_changed()
            else:
                # Inform self and children about loss of a parent
                self.set_changed()

        return removed

    def remove_parent(self, parent, set_changed=True):
        """
        Remove parent from self.parents
        """
        return self.remove_member(parent, child=False, set_changed=set_changed)

    def remove_child(self, child, set_changed=True):
        """
        Remove child from self.children
        """
        return self.remove_member(child, child=True, set_changed=set_changed)

    def set_member(self, member, child=True, set_changed=True):
        """
        Replace all children or parents with `member`.
        """
        if child:
            members = self.children
        else:
            members = self.parents

        # Remove all members
        for old_member in members:
            self.remove_member(old_member, child=child, set_changed=set_changed)

        return self.add_member(member, child=child, set_changed=set_changed)

    def set_parent(self, parent, set_changed=True):
        """
        Replace all parents with `parent`.
        """
        return self.set_member(parent, child=False, set_changed=set_changed)

    def set_child(self, child, set_changed=True):
        """
        Replace all children with `child`.
        """
        return self.set_member(child, child=True, set_changed=set_changed)

    def first_ancestor_instance(self, instance_class, dft=True, level=-1):
        ancestors = self.members(instance_class=instance_class,
                                 descendants=False, includeself=False, dft=dft,
                                 level=level)
        return next(ancestors, None)

    def parent_instances(self, instance_class):
        members = self.members(instance_class=instance_class,
                               descendants=False, includeself=False, level=1)
        return members

    def child_instances(self, instance_class=None):
        members = self.members(instance_class=instance_class,
                               descendants=True, includeself=False, level=1)
        return members

    @property
    def parents(self):
        return self.members(descendants=False, includeself=False, level=1)

    @property
    def children(self):
        return self.members(descendants=True, includeself=False, level=1)

    @property
    def parent(self):
        """
        Return first parent or None.
        """
        return next(self.parents, None)

    @property
    def child(self):
        """
        Return first child or None.
        """
        return next(self.children, None)

    @property
    def max_parents(self):
        return self._node.max_parents

    @max_parents.setter
    def max_parents(self, max_parents):
        self._node.max_parents = max_parents

    @property
    def max_children(self):
        return self._node.max_children

    @max_children.setter
    def max_children(self, max_children):
        self._node.max_children = max_children

    def __str__(self):
        if self.name is None or self.group is None:
            return self.__repr__()
        else:
            return "".join(["".join(self.name).ljust(22),
                            "".join(self.group).ljust(18),
                            "".join(["<", self.__class__.__name__,
                                     ">"]).ljust(14)
                            ])


def _get_node(cargo):
    node = cargo
    if not isinstance(cargo, Node):
        try:
            node = cargo._node
        except:
            raise AttributeError("The cargo does not have an attribute `_node`"
                                 " of type Node, which is necessary to be used"
                                 " as a cargo from an instance of class Node.")
    return node



# -*- coding: utf-8 -*-
"""
author: steve simmert
bugs and hints to: steve.simmert@uni-tuebingen.de
"""

names = 'axis_names'

laterality = 'laterality'

height = 'height'  # sample height
height_unit = 'height_unit'

N_avg = 'number_of_averages'
f_sample = 'sampling_rate'

density = density_p = 'density_particle'
density_p_unit = 'density_particle_unit'

density_m = 'density_medium'
density_m_unit = 'density_medium_unit'

viscosity = 'viscosity'
viscosity_unit = 'viscosity_unit'

material = 'material'

medium = 'medium'

temp = 'temperature'
temp_err = 'temperature_err'
temp_unit = temp_unit = 'temperature_unit'

bead_dia = 'bead_diameter'
bead_dia_err = 'bead_diameter_err'
bead_dia_unit = 'bead_diameter_unit'

radius = 'radius'
radius_err = 'radius_err'
radius_unit = 'radius_unit'

freq = 'freq'
freq_unit = 'frequency_unit'

PSD_x = 'PSD_x'
PSD_y = 'PSD_y'
PSD_z = 'PSD_z'
psd_unit = 'psd_unit'

ex_axis = 'excitation_axis'
ex_freq = 'excitation_frequency'
ex_amp = 'excitation_amplitude'
ex_amp_err = 'excitation_amplitude_err'
ex_amp_unit = 'excitation_amplitude_unit'
ex_pow = 'reference power'
ex_pow_err = 'reference_power_err'
ex_pow_unit = 'reference_power_unit'

z_signal = 'z_signal'

wl = 'wavelength'  # optical wavelength (in air) of the trapping laser
wl_unit = 'wavelength_unit'

fs = focal_shift = 'focal_shift'
dfs = 'focal_shift_err'

ref_ind = 'refractive_index'  # refractive index of the medium

conf_level = 'confidence_level'

fit_method = 'fit_method'

corr = 'correction_factor'
corr_err = 'correction_factor_error'

d_app_surf = 'apparent_surface_distance'
d_app_surf_err = 'apparent_surface_distance_err'
h0 = 'h0'
h0_err = 'h0_err'
height_offset = 'applied_height_offset'

dissens = 'displacement_sensitivity'
dissens_err = 'displacement_sensitivity_error'
dissens_unit = beta_unit = 'displacement_sensitivity_unit'

trap_stiffness = 'trap_stiffness'
trap_stiffness_err = 'trap_stiffness_error'
trap_stiffness_unit = 'trap_stiffness_unit'
kappa_unit = 'trapstiffness_unit'

drag = 'drag_coefficient'
drag_err = 'drag_error'
drag_unit = 'drag_unit'

f_c = 'corner_frequency'
df_c = 'corner_frequency_err'
f_c_unit = 'corner_frequency_unit'

D = 'Diffusion_coefficient'
dD = 'Diffusion_coefficient_err'
D_unit = 'Diffusion_coefficient_unit'

model = 'model_name'
redchi2 = 'reduced_chi2'
nfree = 'number_degrees_freedom'
outlier = 'outlier'
conf_level = 'confidence_level'
bounds = 'freq_bounds'



# -*- coding: utf-8 -*-
# """
# - Author: steve simmert
# - E-mail: steve.simmert@uni-tuebingen.de
# - Copyright: 2015
# """
"""
Manage the reading and storing of data.
"""

from collections import OrderedDict

from scipy import genfromtxt
from scipy import array

import configparser

from . import name_constants as co


def read_std_data_file(path, lower_names=False, **kwargs):
    '''
    Read a standard data file.

    The file is expected to have the names of the data in the first row. The
    data should be delimited by '\t' (=TABS).

    The output is a dictionary with the names as keys and their respective
    columns as values.
    '''
    data = genfromtxt(path,
                      delimiter='\t',
                      names=True,
                      **kwargs
                      )
    if lower_names:
        names = (name.lower() for name in data.dtype.names)
    else:
        names = data.dtype.names
    cols = array([a for a in zip(*list(data))])

    d = OrderedDict()
    for i, name in enumerate(names):
        d.update({name: cols[i]})

    return d


def read_PSD_parameter_file(path):
    """
    Read a PSD parameter file.

    The parameter file is a chaperone to the PSD data
    file. The file is expected to be in the config file standard.

    The parameters are expected to be in the default section.
    """
    pars = configparser.ConfigParser()

    if not pars.read(path):
        raise Exception('Parameter file could not be read at: {}.'
                        ''.format(path))

    defs = pars.defaults()

    if 'ACTIVE_CALIBRATION' in pars:
        defs.update(pars['ACTIVE_CALIBRATION'])
        defs['active_calibration'] = True
    else:
        defs['active_calibration'] = False

    strings = [co.material, co.ex_axis, 'active_calibration',
               co.names, co.N_avg, co.f_sample, co.medium]

    for k in list(defs):
        if k not in strings and not(k.endswith('unit')):
            defs[k] = float(defs[k])

    return defs


def save_psd_data(path, freq, psd_dict):
    """
    Write psd data to a tab-separated data file.

    Arguments
    ---------
    path : str
        Path to the file.
    freq : array-like
        Frequency vector.
    psd_dict : dict
        Dictionary that holds the data in arrays for each frequency value. The
        Keys are the names of the axes.
    """
    data = OrderedDict()
    data['freq'] = freq
    data.update(psd_dict)
    names = [name for name in data.keys()]

    vals = array([a for a in data.values()]).transpose()

    with open(path, 'w') as fl:
        fl.write('\t'.join(names) + '\n')

        for row in range(len(vals)):
            s = ['{0:1.5E}'.format(si) for si in vals[row]]
            fl.write('\t'.join(s) + '\n')


def save_psd_params(path, param_dict, ac_param_dict=None):
    """
    Saves parameters to a configuration file.

    All key-value pairs in param_dict are written to the 'DEFAULT' section.
    If ac_param_dict dictionary is provided, the key-value pairs are stored
    to the section 'active_calibration'.
    """
    pars = configparser.ConfigParser()
    pars['DEFAULT'] = param_dict

    if ac_param_dict is not None:
        pars['active_calibration'] = ac_param_dict

    with open(path, 'w') as fl:
        pars.write(fl)



# -*- coding: utf-8 -*-
# """
# - Author: steve simmert
# - E-mail: steve.simmert@uni-tuebingen.de
# - Copyright: 2015
# """

from . import ureg

from collections import Iterable

from inspect import getargspec

from lmfit import Parameters

from numpy.lib.scimath import log10
from numpy import logspace


def gen_fit_pars(**kwargs):
    """
    Generate fit parameters.

    Every keyword will create a parameter, keywords ending with '_min' or
    '_max' will define the upper and lower bound for the parameter.

    Example
    -------
    gen_fit_pars(D=1e-6, D_min=0, f_c=1000, f_c_max=1e6) will create
    OrderedDict([('D', <Parameter 'D', 1e-06, bounds=[0:None]>),
    ('f_c', <Parameter 'f_c', 1000, bounds=[None:1000000.0]>)]).

    Returns
    -------
    params : Parameters
        lmfit Parameters object
    """
    params = Parameters()

    for k, v in kwargs.items():
        if k.endswith('_min') or k.endswith('_max'):
            continue

        try:
            v_min = kwargs[k + '_min']
        except:
            v_min = None
        try:
            v_max = kwargs[k + '_max']
        except:
            v_max = None

        params.add(k, value=v, min=v_min, max=v_max)
    return params


def flatten_list(l):
    """
    Flatten an irregular list of lists.

    Found here:
    ('http://stackoverflow.com/questions/2158395/
      flatten-an-irregular-list-of-lists-in-python')
    """
    for el in l:
        if isinstance(el, Iterable) and not isinstance(el, (str, bytes)):
            for sub in flatten_list(el):
                yield sub
        else:
            yield el


def str2u(s):
    """
    Convert a unit represented as a string into a regular string expression.
    """
    return str(ureg(s).units)


def u2str(unit, formatter='P'):
    """ format a pint unit to a pretty print string"""
    if formatter.lower() == 'p':
        return ''.join('{0:P~}'.format(ureg(unit)).split(' ')[1:])
    elif formatter.lower() == 'l':
        return ''.join('{0:L~}'.format(ureg(unit)).split(' ')[1:])
    else:
        raise ValueError(formatter)


def check_and_separate(fun, kws, exclude=True):
    """
    Checks if keywords of a given dictionary match the signature of
    the given function and separates the keywords.
    """
    argspec = getargspec(fun)

    sep = {}
    for k in kws:
        if k in argspec.args:
            sep[k] = kws[k]

    if exclude:
        for k in sep:
            kws.pop(k)

    return sep


def logspace_points_per_decade(start, end, ppd=5):
    """
    Returns a vector with a fixed number of points per decade.

    When a provided with a range of full decades, like [1, 100] the resulting
    numbers will only differ by a decimal shift. The end will be included.

    Arguments
    ---------
    start : float
        Stating number, e.g. 1.0
    end : float
        last number of the vector, e.g. 1e5
    ppd : int
        Number of points per decade.
    """
    ndec = log10(end) - log10(start)
    num = int(ndec * ppd - (ndec-1))
    return logspace(log10(start), log10(end), num=num)



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 11 14:37:58 2016

@author: Tobias Jachowski
"""
import numpy as np
import struct
import os

from pyoti.data.datasource import DataSource


class CNLabViewBinData(DataSource):
    def __init__(self, filename, directory=None, parafile=None, datext='.bin',
                 parext='_para.dat', **kwargs):
        """
        parext : str, optional
            The extension of the parameter file ('_para.dat', default)
        """
        super().__init__(filename=filename, directory=directory, **kwargs)

        if parafile is None:
            parafile = self.absfile.replace(datext, parext)
        self._parafile_orig = parafile

        self.samplingrate = get_samplingrate_from_parameter_file(parafile)

        self.name = ("LabVIEW bin data originally loaded from \n"
                     "    %s with \n"
                     "    samplingrate %s Hz") % (self.absfile_orig,
                                                  self.samplingrate)

    def as_array(self):
        filename = self.absfile
        data = read_labview_bin_data(filename)
        return data

    @property
    def parafile_orig(self):
        return self._parafile_orig


class CNLabViewTxtData(DataSource):
    def __init__(self, filename, directory=None, samplingrate=1000.0,
                 **kwargs):
        """
        parext : str, optional
            The extension of the parameter file ('_para.dat', default)
        """
        super().__init__(filename=filename, directory=directory, **kwargs)
        self.samplingrate = samplingrate

        self.name = ("LabVIEW txt data originally loaded from \n"
                     "    %s with \n"
                     "    samplingrate %s Hz") % (self.absfile_orig,
                                                  self.samplingrate)

    def as_array(self):
        filename = self.absfile
        data = np.loadtxt(filename, skiprows=5)
        return data


class CNLabViewQPDviData(CNLabViewTxtData):
    def __init__(self, filename, directory=None, **kwargs):
        super().__init__(filename, directory=directory, **kwargs)
        self.samplingrate = np.genfromtxt(self.absfile, skip_header=2,
                                          max_rows=1)[0]

        self.name = ("LabVIEW QPD.vi data originally loaded from \n"
                     "    %s with \n"
                     "    samplingrate %s Hz") % (os.path.join(self.directory,
                                                               self.filename),
                                                  self.samplingrate)


def get_samplingrate_from_parameter_file(parafile):
    para = np.loadtxt(parafile, comments='%', delimiter='\t')
    scanrate = para[0]  # float
    decimating = para[2]  # float
    samplingrate = scanrate / decimating  # float
    return samplingrate


def chunk_info(filename):
    """
    Read all chunk arrays with the information (number of rows and columns) of
    the data, following each chunk.
    """
    with open(filename, "rb") as f:
        pos = 0
        while True:
            byte = f.read(8)
            pos += 1
            if byte:
                chunk_shape = struct.unpack('>2i', byte)
                if chunk_shape[0] * chunk_shape[1] > 0:
                    # go to byte offset relative to current strem position
                    f.seek(8 * chunk_shape[0] * chunk_shape[1], 1)
                    yield(pos, *chunk_shape)
                    pos += chunk_shape[0] * chunk_shape[1]
            else:
                break


def chunk_data(filename, chunks, dtype='>d'):
    with open(filename, "rb") as f:
        for chunk in chunks:
            # go to byte offset where chunk starts
            f.seek(chunk[0] * 8)
            # read the data
            data_bin = f.read(chunk[1] * chunk[2] * 8)
            data = np.fromstring(data_bin, dtype=dtype)
            yield data


def read_labview_bin_data(filename, dtype='>d', start_row_idx=0,
                          number_of_rows=-1):
    """
    Parameters
    ----------
    filename : str
        Path of the binary file to read
    dtype : str
        Type of double
    start_row_idx : int
        Index of the first datapoint (of all traces) to read
    number_or_rows : int
        Number of datapoints (of all traces) to read. Defaults to number of
        datapoints of the binary file - `start_row_idx`
    """
    stop_row_idx = start_row_idx + number_of_rows  # index of row to stop read
    rows_to_read = 0  # rows to read after getting information of chunks
    columns = None  # number of columns (i.e. traces) in binary file
    chunks = []  # chunks with the index information of data to be read
    chunk_row_start = 0  # running index of first row of chunk
    chunk_row_stop = 0  # running stop index of last row of chunk

    if number_of_rows == 0:
        return np.empty((0, 0))

    print('Getting chunk info from:')
    print('  \'%s\'' % filename)
    for chunk in chunk_info(filename):
        # Set the running indices to the new position (row)
        chunk_row_start = chunk_row_stop
        chunk_row_stop += chunk[1]
        # Check if information about number of columns changed from one to the
        # next chunk
        if columns == chunk[2] or columns is None:
            columns = chunk[2]
        else:
            print("Number of columns in chunks of file differ from each other!")
        # Check if chunk is (partly) contained within the requested data index
        if (chunk_row_stop > start_row_idx
                and (number_of_rows < 0 or chunk_row_start < stop_row_idx)):
            # Check read position and number of rows of first chunk
            if len(chunks) is 0:
                shift = start_row_idx - chunk_row_start
                chunk = (chunk[0] + shift * chunk[2],
                         chunk[1] - shift,
                         chunk[2])
            chunks.append(chunk)
            rows_to_read += chunk[1]
        # Check stop position and number of rows of last chunk
        if (number_of_rows > 0 and chunk_row_stop >= stop_row_idx):
            shift = chunk_row_stop - stop_row_idx
            chunk = (chunk[0], chunk[1] - shift, chunk[2])
            chunks[-1] = chunk
            rows_to_read -= shift
            break

    if rows_to_read == 0:
        return np.empty((0, 0))

    print('Reading data chunks from:')
    print('  \'%s\'' % filename)
    data = np.empty(rows_to_read * columns)
    i = 0
    for _data in chunk_data(filename, chunks, dtype=dtype):
        length = len(_data)
        data[i:i + length] = _data
        i += length

    data = data.reshape(rows_to_read, columns)

    # Workaround, if last column is 0.0
    if np.all(data[:, -1] == 0.0):
        # kick out last column
        data = data[:, :-1]

    return data



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 15 15:29:19 2016

@author: Tobias Jachowski
"""
import matplotlib.pyplot as plt
import numpy as np

from . import signal as sn
from .evaluate import Evaluator
from .signalfeature import StepFinder


class Motion(Evaluator):
    """
    Calculates motion related parameters, like steps.

    To identify positive and negative steps, a Stepfinder is used.
    """
    def __init__(self, region=None, calibration=None, resolution=1000,
                 filter_time=-1.0, min_dwell_time=0.0, shift_time=0.0,
                 impose_dwell_time=0.0, resolution_sf=None,
                 filter_time_sf=None, **kwargs):
        """
        This __init__() constructor extends the superclass (Evaluator)
        constructor by setting the traces_sf to default to 'positionZ'.

        For extensive documentation of the Parameters and additional
        attributes, see `pyoti.evaluate.evaluate.Evaluator`.

        Parameters
        ----------
        region : pyoti.region.region.Region
        calibration : pyoti.calibration.calibration.Calibration
        resolution : float
        filter_time : float
        min_dwell_time : float
            Minimum time a step plateau needs to have.
        shift_time : float
            Time the start of the plateau is shifted.
        impose_dwell_time : float
            Adjust the stop index of the returned detected plateaus such a way,
            that they have a length of the given min_dwell_time.
        resolution_sf : float
        filter_time_sf : float
        **kwargs
            Extra arguments, e.g. traces_sf, in case the positionXY signals are
            called differently.

        Attributes
        ----------
        min_dwell_time
        min_plateau_length
        shift_time
        shift_length
        impose_dwell_time
        impose_plateau_length
        include_first_last_plateau
        pos_steps
        neg_steps
        steps
        pos_plateaus
        neg_plateaus
        plateaus
        """
        # resolution for (almost) all properties
        # resolution 1.000 Hz, samplingrate 40.000 Hz -> 40 points

        traces_sf = kwargs.pop('traces_sf', 'positionZ')

        super().__init__(region=region, calibration=calibration,
                         resolution=resolution, filter_time=filter_time,
                         sf_class=StepFinder, traces_sf=traces_sf,
                         resolution_sf=resolution_sf,
                         filter_time_sf=filter_time_sf)

        self._min_dwell_time = min_dwell_time  # s
        self._shift_time = shift_time  # s
        self._impose_dwell_time = impose_dwell_time  # s

    def _sections(self, plateaus=True, direction=None, **kwargs):
        if plateaus:
            if direction == 'pos':
                sections = self._sf.pos_plateau_segments
            elif direction == 'neg':
                sections = self._sf.neg_plateau_segments
            else:
                sections = self._sf.plateau_segments
            sections = self._shift_and_size_segments(sections)
        else:  # steps instead of plateaus
            if direction == 'pos':
                steps = self._sf.pos_steps
            elif direction == 'neg':
                steps = self._sf.neg_steps
            else:
                steps = self._sf.steps
            steps = steps + self.shift_length
            steps = sn.limit_segments(steps, min_start=0,
                                      max_stop=self.region.datapoints)
            sections = steps

        return sections

    def _shift_and_size_segments(self, segs):

        if self.impose_dwell_time > 0.0:
            # Select the plateaus having at least a size of min_dwell_time or
            # shift_time and impose_dwell_time
            size = max(self.min_plateau_length, self.shift_length
                       + self.impose_plateau_length)
            segs = sn.get_min_size_segments(segs, size)
            # Shift and size the plateaus and eventually crop the first/last
            # plateaus to min_start and max_stop
            segs = sn.shift_and_size_segments(segs, shift=self.shift_length,
                                              size=self.impose_plateau_length,
                                              min_start=0,
                                              max_stop=self.region.datapoints)
        else:
            # Select the plateaus having at least a size of min_dwell_time
            segs = sn.get_min_size_segments(segs, self.min_plateau_length)
            # Shift the plateaus and eventually crop the first/last plateaus to
            # min_start and max_stop
            segs = sn.shift_segments(segs, shift=self.shift_length,
                                     min_start=0,
                                     max_stop=self.region.datapoints)

        # After eventually having cropped the segments, reselect the plateaus
        # having at least a length of min_dwell_time
        segs = sn.get_min_size_segments(segs, self.min_plateau_length)

        return segs

    def _rfigure(self, legend=True):
        # Create figure and close old one
        figure = plt.figure()
        ax = figure.gca()

        ax.grid(True)

        # Get t, positionZ, psdXYZ and detected plateaus
        t = self.timevector
        positionZ = self.get_data(traces=self.traces_sf)
        plateaus = self.plateaus

        # Plot the original signal
        ax.plot(t, positionZ * 1e6, color='k', alpha=0.1)

        # Plot the detected plateaus (positionZ) and store psd data chunks and
        # heights
        for plat in plateaus:
            ax.plot(t[plat], positionZ[plat] * 1e6)

        ax.set_xlabel('Time (s)')
        ax.set_ylabel("Signal %s (um)" % (self.traces_sf))
        figure.suptitle('Automatically detected plateaus.')

        return figure

    @property
    def min_dwell_time(self):
        return self._min_dwell_time

    @min_dwell_time.setter
    def min_dwell_time(self, min_dwell_time):
        self._min_dwell_time = min_dwell_time
        self.update()

    @property
    def min_plateau_length(self):
        return int(np.round(self.min_dwell_time * self.resolution))

    @property
    def shift_time(self):
        return self._shift_time

    @shift_time.setter
    def shift_time(self, shift_time):
        self._shift_time = shift_time
        self.update()

    @property
    def shift_length(self):
        return int(np.round(self.shift_time * self.resolution))

    @property
    def impose_dwell_time(self):
        return self._impose_dwell_time

    @impose_dwell_time.setter
    def impose_dwell_time(self, impose_dwell_time):
        self._impose_dwell_time = impose_dwell_time
        self.update()

    @property
    def impose_plateau_length(self):
        return int(np.round(self.impose_dwell_time * self.resolution))

    @property
    def include_first_last_plateau(self):
        return self._sf.include_first_last

    @include_first_last_plateau.setter
    def include_first_last_plateau(self, include):
        self._sf.include_first_last = True
        self.update()

    @property
    def pos_steps(self):
        return self.sections(plateaus=False, direction='pos')

    @property
    def neg_steps(self):
        return self.sections(plateaus=False, direction='neg')

    @property
    def steps(self):
        return self.sections(plateaus=False)

    @property
    def pos_plateaus(self):
        return self.sections(direction='pos')

    @property
    def neg_plateaus(self):
        return self.sections(direction='neg')

    @property
    def plateaus(self):
        return self.sections()



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 18 13:41:17 2016

@author: Tobias Jachowski
"""
import inspect
import numbers

from pyoti.data.datasource import DataSource
from pyoti.picklable import unboundfunction


class GenericDataFile(DataSource):
    def __init__(self, load_data, filename, directory=None, samplingrate=1.0,
                 **kwargs):
        """
        load_data : function
        filename : str
        directory : str
        samplingrate : float
        **kwargs
        """
        super().__init__(filename=filename, directory=directory, **kwargs)

        self.load_data = unboundfunction(load_data)

        if isinstance(samplingrate, numbers.Number):
            self.samplingrate = samplingrate
        else:
            samplingrate_args = {}
            for par in inspect.getargspec(samplingrate)[0]:
                if par in kwargs:  # par can be anything, except load_data,
                                   # filename, directory or samplingrate
                    samplingrate_args[par] = kwargs.get(par)
                if par == 'filename':  # automatically use filename
                    samplingrate_args['filename'] = self.absfile
            self.samplingrate = samplingrate(**samplingrate_args)

        self.load_data_args = {}
        for par in inspect.getargspec(load_data)[0]:
            if par in kwargs:
                self.load_data_args[par] = kwargs.get(par)

        self.name = ("Generic data originally loaded from \n"
                     "    %s with \n"
                     "    samplingrate %s Hz") % (self.absfile_orig,
                                                  self.samplingrate)

    def as_array(self):
        filename = self.absfile
        data = self.load_data(filename, **self.load_data_args)
        return data


class GenericData(DataSource):
    def __init__(self, load_data, samplingrate=1.0, **kwargs):
        """
        load_data : function
        samplingrate : float
        """
        self.load_data = unboundfunction(load_data)

        if isinstance(samplingrate, numbers.Number):
            self.samplingrate = samplingrate
        else:
            samplingrate_args = {}
            for par in inspect.getargspec(samplingrate)[0]:
                if par in kwargs:
                    samplingrate_args[par] = kwargs.pop(par)
            self.samplingrate = samplingrate(**samplingrate_args)

        self.fun_args = {}
        for par in inspect.getargspec(load_data)[0]:
            if par in kwargs:
                self.fun_args[par] = kwargs.pop(par)

        self.name = ("Generic data with \n"
                     "    samplingrate %s Hz") % (self.samplingrate)

    def as_array(self):
        data = self.load_data(**self.fun_args)
        return data



# -*- coding: utf-8 -*-
"""
Created on Thu Dec  3 13:39:13 2015

@author: Tobias Jachowski
"""
import collections
import sys

from . import config as cf
from . import helpers as hp


class Traces(object):

    def __init__(self, cfgfile, **kwargs):
        self._cfgfile = cfgfile
        self.reload()

    def __getattr__(self, name):
        return self.normalize(name)

    def __contains__(self, key):
        return key in self._TRACE_ALIASES

    def reload(self, cfgfile=None):
        cfgfile = cfgfile or self.cfgfile
        cfg = cf.read_cfg_file(cfgfile)
        self._cfgfile = cfgfile

        self._TRACE_ALIASES = {a: [t.strip() for t in traces.split(',')]
                               for a, traces
                               in cfg['aliases'].items()}

        self._FACTOR = {t: float(f)
                        for t, f
                        in cfg['factor'].items()}

        self._COLOR = {t: c
                       for t, c
                       in cfg['color'].items()}

        self._LABEL = {t: l
                       for t, l
                       in cfg['label'].items()}

    def normalize(self, traces):
        """
        Normalize traces returns a list of trace names.
        It looks up alias names for traces and replaces them with the list of
        names. If alias is not found, the input parameters are converted /
        added to the return list.
        """
        traces = hp.listify(traces)

        if isinstance(traces, collections.Iterable):
            _traces = []
            for trace in traces:
                if isinstance(trace, str) and trace in self._TRACE_ALIASES:
                    # lookup for traces as alias, a shorthand notation, or a
                    # combination
                    trace = self._TRACE_ALIASES[trace]
                    _traces.extend(trace)
                else:
                    _traces.append(trace)

            # Remove duplicate entries
            seen = set()
            seen_add = seen.add
            traces = [trace
                      for trace in _traces
                      if not (trace in seen or seen_add(trace))]

        return traces

    def factor(self, trace):
        if trace in self._FACTOR:
            return self._FACTOR[trace]
        elif 'DEFAULT' in self._FACTOR:
            return self._FACTOR['DEFAULT']
        else:
            return 1

    def label(self, trace):
        if trace in self._LABEL:
            return self._LABEL[trace]
        elif 'DEFAULT' in self._LABEL:
            return self._LABEL['DEFAULT']
        else:
            return 'Data (arbitrary)'

    def color(self, trace):
        if trace in self._COLOR:
            return self._COLOR[trace]
        elif 'DEFAULT' in self._COLOR:
            return self._COLOR['DEFAULT']
        else:
            return 'black'

    @property
    def cfgfile(self):
        return self._cfgfile

    @cfgfile.setter
    def cfgfile(self, cfgfile):
        self.reload(cfgfile=cfgfile)

    # ### module like functionality ###
    @property
    def __file__(self):
        return __file__

    @property
    def __name__(self):
        return __name__

    @property
    def __package__(self):
        return __package__

cfgfile='traces.cfg'
sys.modules[__name__] = Traces(cfgfile)
